# urls.py
# Path: appointment/urls.py

"""
Author: Adams Pierre David
Since: 1.0.0
"""

from django.urls import include, path

from appointment.views import (
    appointment_client_information, appointment_request, appointment_request_submit, confirm_reschedule,
    default_thank_you, enter_verification_code, get_available_slots_ajax, get_next_available_date_ajax,
    get_non_working_days_ajax, prepare_reschedule_appointment, reschedule_appointment_submit, set_passwd
)
from appointment.views_admin import (
    add_day_off, add_or_update_service, add_or_update_staff_info, add_staff_member_info, add_working_hours,
    create_new_staff_member, delete_appointment, delete_appointment_ajax, delete_day_off, delete_service,
    delete_working_hours, display_appointment, email_change_verification_code, fetch_service_list_for_staff,
    fetch_staff_list, get_service_list, get_user_appointments, is_user_staff_admin, make_superuser_staff_member,
    remove_staff_member, remove_superuser_staff_member, update_appt_date_time, update_appt_min_info, update_day_off,
    update_personal_info, update_working_hours, user_profile, validate_appointment_date
)

app_name = 'appointment'

admin_urlpatterns = [
    # display the calendar with the events
    path('appointments/<str:response_type>/', get_user_appointments, name='get_user_event_type'),
    path('appointments/', get_user_appointments, name='get_user_appointments'),

    # create a new staff member and make/remove superuser staff member
    path('add-staff-member-info/', add_staff_member_info, name='add_staff_member_info'),
    path('create-new-staff-member/', create_new_staff_member, name='add_staff_member_personal_info'),
    path('update-staff-member/<int:user_id>/', add_or_update_staff_info, name='update_staff_other_info'),
    path('add-staff-member/', add_or_update_staff_info, name='add_staff_other_info'),
    path('make-superuser-staff-member/', make_superuser_staff_member, name='make_superuser_staff_member'),
    path('remove-superuser-staff-member/', remove_superuser_staff_member, name='remove_superuser_staff_member'),

    # remove staff member
    path('remove-staff-member/<int:staff_user_id>/', remove_staff_member, name='remove_staff_member'),

    # add, update, remove services
    path('add-service/', add_or_update_service, name='add_service'),
    path('update-service/<int:service_id>/', add_or_update_service, name='update_service'),
    path('delete-service/<int:service_id>/', delete_service, name='delete_service'),
    path('service-list/', get_service_list, name='get_service_list'),
    path('service-list/<str:response_type>/', get_service_list, name='get_service_list_type'),
    path('view-service/<int:service_id>/<int:view>/', add_or_update_service, name='view_service'),

    # display details for one event
    path('display-appointment/<int:appointment_id>/', display_appointment, name='display_appointment'),

    # complete profile
    path('user-profile/<int:staff_user_id>/', user_profile, name='user_profile'),
    path('user-profile/', user_profile, name='user_profile'),
    path('update-user-info/<int:staff_user_id>/', update_personal_info, name='update_user_info'),
    path('update-user-info/', update_personal_info, name='update_user_info'),

    # add, update, delete day off with staff_user_id
    path('add-day-off/<int:staff_user_id>/', add_day_off, name='add_day_off'),
    path('update-day-off/<int:day_off_id>/<int:staff_user_id>/', update_day_off, name='update_day_off_id'),
    path('delete-day-off/<int:day_off_id>/<int:staff_user_id>/', delete_day_off, name='delete_day_off_id'),

    # add, update, delete day off without staff_user_id
    path('update-day-off/<int:day_off_id>/', update_day_off, name='update_day_off'),
    path('delete-day-off/<int:day_off_id>/', delete_day_off, name='delete_day_off'),

    # add, update, delete working hours with staff_user_id
    path('update-working-hours/<int:working_hours_id>/<int:staff_user_id>/', update_working_hours,
         name='update_working_hours_id'),
    path('add-working-hours/<int:staff_user_id>/', add_working_hours, name='add_working_hours_id'),
    path('delete-working-hours/<int:working_hours_id>/<int:staff_user_id>/', delete_working_hours,
         name='delete_working_hours_id'),

    # add, update, delete working hours without staff_user_id
    path('update-working-hours/<int:working_hours_id>/', update_working_hours, name='update_working_hours'),
    path('add-working-hours/', add_working_hours, name='add_working_hours'),
    path('delete-working-hours/<int:working_hours_id>/', delete_working_hours, name='delete_working_hours'),

    # delete appointment
    path('delete-appointment/<int:appointment_id>/', delete_appointment, name='delete_appointment'),
]

ajax_urlpatterns = [
    path('available_slots/', get_available_slots_ajax, name='available_slots_ajax'),
    path('request_next_available_slot/<int:service_id>/', get_next_available_date_ajax,
         name='request_next_available_slot'),
    path('request_staff_info/', get_non_working_days_ajax, name='get_non_working_days_ajax'),
    path('fetch_service_list_for_staff/', fetch_service_list_for_staff, name='fetch_service_list_for_staff'),
    path('fetch_staff_list/', fetch_staff_list, name='fetch_staff_list'),
    path('update_appt_min_info/', update_appt_min_info, name="update_appt_min_info"),
    path('update_appt_date_time/', update_appt_date_time, name="update_appt_date_time"),
    path('validate_appointment_date/', validate_appointment_date, name="validate_appointment_date"),
    # delete appointment ajax
    path('delete_appointment/', delete_appointment_ajax, name="delete_appointment_ajax"),
    path('is_user_staff_admin/', is_user_staff_admin, name="is_user_staff_admin"),
]

urlpatterns = [
    # homepage
    path('request/<int:service_id>/', appointment_request, name='appointment_request'),
    path('request-submit/', appointment_request_submit, name='appointment_request_submit'),
    path('appointment/<str:id_request>/reschedule/', prepare_reschedule_appointment,
         name='prepare_reschedule_appointment'),
    path('appointment-reschedule-submit/', reschedule_appointment_submit, name='reschedule_appointment_submit'),
    path('confirm-reschedule/<str:id_request>/', confirm_reschedule, name='confirm_reschedule'),
    path('client-info/<int:appointment_request_id>/<str:id_request>/', appointment_client_information,
         name='appointment_client_information'),
    path('verification-code/<int:appointment_request_id>/<str:id_request>/', enter_verification_code,
         name='enter_verification_code'),
    path('verification-code/', email_change_verification_code, name='email_change_verification_code'),
    path('thank-you/<int:appointment_id>/', default_thank_you, name='default_thank_you'),
    path('verify/<uidb64>/<str:token>/', set_passwd, name='set_passwd'),
    path('ajax/', include(ajax_urlpatterns)),
    path('app-admin/', include(admin_urlpatterns)),
]



# tasks.py
# Path: appointment/tasks.py

"""
Author: Adams Pierre David
Since: 3.1.0
"""
from django.utils.translation import gettext as _

from appointment.email_sender import notify_admin, send_email
from appointment.logger_config import get_logger
from appointment.models import Appointment

logger = get_logger(__name__)


def send_email_reminder(to_email, first_name, reschedule_link, appointment_id):
    """
    Send a reminder email to the client about the upcoming appointment.
    """

    # Fetch the appointment using appointment_id
    logger.info(f"Sending reminder to {to_email} for appointment {appointment_id}")
    appointment = Appointment.objects.get(id=appointment_id)
    recipient_type = 'client'
    email_context = {
        'first_name': first_name,
        'appointment': appointment,
        'reschedule_link': reschedule_link,
        'recipient_type': recipient_type,
    }
    send_email(
            recipient_list=[to_email], subject=_("Reminder: Upcoming Appointment"),
            template_url='email_sender/reminder_email.html', context=email_context
    )
    # Notify the admin
    logger.info(f"Sending admin reminder also")
    email_context['recipient_type'] = 'admin'
    notify_admin(
            subject=_("Admin Reminder: Upcoming Appointment"),
            template_url='email_sender/reminder_email.html', context=email_context
    )


def send_email_task(recipient_list, subject, message, html_message, from_email):
    """
   Task function to send an email asynchronously using Django's send_mail function.
   This function tries to send an email and logs an error if it fails.
   """
    try:
        from django.core.mail import send_mail
        logger.info(f"Sending email to {recipient_list} with subject: {subject}")
        send_mail(
                subject=subject, message=message, html_message=html_message, from_email=from_email,
                recipient_list=recipient_list, fail_silently=False,
        )
    except Exception as e:
        logger.error(f"Error sending email from task: {e}")


def notify_admin_task(subject, message, html_message):
    """
    Task function to send an admin email asynchronously.
    """
    try:
        from django.core.mail import mail_admins
        logger.info(f"Sending admin email with subject: {subject}")
        mail_admins(subject=subject, message=message, html_message=html_message, fail_silently=False)
    except Exception as e:
        logger.error(f"Error sending admin email from task: {e}")



# settings.py
# Path: appointment/settings.py

"""
Author: Adams Pierre David
Since: 1.0.0
"""

from django.conf import settings
from django.conf.global_settings import DEFAULT_FROM_EMAIL

from appointment.logger_config import get_logger

logger = get_logger(__name__)

APPOINTMENT_BASE_TEMPLATE = getattr(settings, 'APPOINTMENT_BASE_TEMPLATE', 'base_templates/base.html')
APPOINTMENT_ADMIN_BASE_TEMPLATE = getattr(settings, 'APPOINTMENT_ADMIN_BASE_TEMPLATE', 'base_templates/base.html')
APPOINTMENT_WEBSITE_NAME = getattr(settings, 'APPOINTMENT_WEBSITE_NAME', 'Website')
APPOINTMENT_PAYMENT_URL = getattr(settings, 'APPOINTMENT_PAYMENT_URL', None)
APPOINTMENT_THANK_YOU_URL = getattr(settings, 'APPOINTMENT_THANK_YOU_URL', None)
APPOINTMENT_SLOT_DURATION = getattr(settings, 'APPOINTMENT_SLOT_DURATION', 30)
APPOINTMENT_BUFFER_TIME = getattr(settings, 'APPOINTMENT_BUFFER_TIME', 0)
APPOINTMENT_LEAD_TIME = getattr(settings, 'APPOINTMENT_LEAD_TIME', (9, 0))
APPOINTMENT_FINISH_TIME = getattr(settings, 'APPOINTMENT_FINISH_TIME', (18, 30))
APP_DEFAULT_FROM_EMAIL = getattr(settings, 'DEFAULT_FROM_EMAIL', DEFAULT_FROM_EMAIL)


def check_q_cluster():
    """
    Checks if Django Q is properly installed and configured in the Django settings.
    If 'django_q' is not in INSTALLED_APPS, it warns about both 'django_q' not being installed
    and 'Q_CLUSTER' likely not being configured.
    If 'django_q' is installed but 'Q_CLUSTER' is not configured, it only warns about 'Q_CLUSTER'.
    Returns True if configurations are correct, otherwise False.
    """
    missing_conf = []
    logger.info("Checking missing configuration for django q cluster")
    # Check if Django Q is installed
    if 'django_q' not in settings.INSTALLED_APPS:
        missing_conf.append("Django Q is not in settings.INSTALLED_APPS. Please add it to the list. "
                            "See https://django-appt-doc.adamspierredavid.com/getting-started/#installation "
                            "for more information")

    # Check if Q_CLUSTER configuration is defined
    if not hasattr(settings, 'Q_CLUSTER'):
        missing_conf.append("Q_CLUSTER is not defined in settings. Please define it. "
                            "See https://django-appt-doc.adamspierredavid.com/project-structure/#configuration "
                            "for more information.")

    # Log warnings if any configurations are missing
    if missing_conf:
        for warning in missing_conf:
            logger.warning(warning)
        return False
    # Both 'django_q' is installed and 'Q_CLUSTER' is configured
    logger.info("Django Q cluster is properly configured")
    return True



from copy import deepcopy
from datetime import timedelta

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import SimpleUploadedFile

from appointment.models import Service
from appointment.tests.base.base_test import BaseTest


class ServiceCreationAndBasicAttributesTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_service_creation(self):
        self.assertIsNotNone(self.service)

    def test_basic_attributes_verification(self):
        self.assertEqual(self.service.name, "Stargate Activation")
        self.assertEqual(self.service.description, "Activate the Stargate")
        self.assertEqual(self.service.duration, timedelta(hours=1))

    def test_timestamps_on_creation(self):
        """Newly created services should have created_at and updated_at values."""
        self.assertIsNotNone(self.service.created_at)
        self.assertIsNotNone(self.service.updated_at)


class ServicePriceTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_service_price_verification(self):
        self.assertEqual(self.service.price, 100000)
        self.assertEqual(self.service.get_price(), 100000)

    def test_paid_service_verification(self):
        self.assertTrue(self.service.is_a_paid_service())

    def check_price(self, price, expected_string):
        service = deepcopy(self.service)
        service.price = price
        self.assertEqual(service.get_price_text(), expected_string)

    def test_dynamic_price_representation(self):
        """Test that the get_price method returns the correct string for a service with a price of 100, 1000, etc."""
        test_cases = [
            (100, '100$'),
            (100.50, '100.5$'),
            (49.99, '49.99$'),
            (0, 'Free')
        ]
        for price, expected in test_cases:
            self.check_price(price, expected)


class ServiceDurationTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_service_duration_verification(self):
        """Test the duration of the service."""
        self.assertEqual(self.service.duration, self.service1.duration)

    def check_duration(self, duration, expected_string):
        service = deepcopy(self.service)
        service.duration = duration
        self.assertEqual(service.get_duration(), expected_string)

    def test_dynamic_duration_representation(self):
        """Test that the get_duration method returns the correct string for a service with a duration of 30 seconds,
        30 minutes, etc."""
        test_cases = [
            (timedelta(seconds=30), '30 seconds'),
            (timedelta(minutes=30), '30 minutes'),
            (timedelta(hours=1), '1 hour'),
            (timedelta(hours=2, minutes=30), '2 hours 30 minutes'),
            (timedelta(days=1), '1 day'),
        ]
        for duration, expected in test_cases:
            self.check_duration(duration, expected)


class ServiceDownPaymentTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_down_payment_value(self):
        """By default, down payment value is 0"""
        self.assertEqual(self.service.down_payment, 0)
        self.assertEqual(self.service.get_down_payment(), 0)

        self.assertEqual(self.service.get_down_payment_text(), "Free")

        # Change the down payment value to 69.99
        s = deepcopy(self.service)
        s.down_payment = 69.99
        self.assertEqual(s.get_down_payment(), 69.99)

        self.assertEqual(s.get_down_payment_text(), "69.99$")

    def test_accepts_down_payment(self):
        """By default, down payment is not accepted."""
        self.assertFalse(self.service.accepts_down_payment())

        # Change the accepts_down_payment value to True
        s = deepcopy(self.service)
        s.down_payment = 69.99
        self.assertTrue(s.accepts_down_payment())

    def test_equal_price_and_down_payment_scenario(self):
        """A service can be created with a price and down payment of the same value.
        This is useful when the service requires full payment upfront.
        """
        service = Service.objects.create(
            name="Naquadah Generator Maintenance", duration=timedelta(hours=1), price=100, down_payment=100)
        self.assertEqual(service.price, service.down_payment)


class ServiceRepresentationAndMiscTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_string_representation_of_service(self):
        """Test the string representation of the Service model."""
        service_name = "Test Service"
        service = Service.objects.create(name=service_name, duration=timedelta(hours=1), price=100)
        self.assertEqual(str(service), service_name)

    def test_image_url_with_attached_image(self):
        """Service should return the correct URL for the image if provided."""
        # Create an image and attach it to the service
        image_path = settings.BASE_DIR / 'appointment/static/img/texture.webp'  # Adjust the path as necessary
        image = SimpleUploadedFile(name='test_image.png', content=open(image_path, 'rb').read(),
                                   content_type='image/png')
        service = Service.objects.create(name="Service with Image", duration=timedelta(hours=1), price=50, image=image)

        # Assuming you have MEDIA_URL set in your settings for development like '/media/'
        expected_url = f"{settings.MEDIA_URL}{service.image}"
        self.assertTrue(service.get_image_url().endswith(expected_url))

    def test_image_url_without_attached_image(self):
        """Service should handle cases where no image is provided gracefully."""
        service = Service.objects.create(name="Gate Travel Coordination", duration=timedelta(hours=1), price=50)
        self.assertEqual(service.get_image_url(), "")

    def test_auto_generation_of_background_color(self):
        """Service should auto-generate a background color if none is provided."""
        service = Service.objects.create(name="Wormhole Stability Analysis", duration=timedelta(hours=1), price=50)
        self.assertIsNotNone(service.background_color)
        self.assertNotEqual(service.background_color, "")

    def test_service_to_dict_representation(self):
        """Test the to_dict method returns the correct dictionary representation of the Service instance."""
        service = Service.objects.create(name="Off-world Tactical Training", duration=timedelta(hours=1), price=150,
                                         description="Train for off-world missions")
        expected_dict = {
            "id": service.id,
            "name": "Off-world Tactical Training",
            "description": "Train for off-world missions",
            "price": "150"
        }
        self.assertEqual(service.to_dict(), expected_dict)

    def test_reschedule_features(self):
        """Service should correctly handle reschedule limits and rescheduling allowance."""
        service = Service.objects.create(name="Goa'uld Artifact Decryption", duration=timedelta(hours=1), price=50,
                                         reschedule_limit=3, allow_rescheduling=True)
        self.assertEqual(service.reschedule_limit, 3)
        self.assertTrue(service.allow_rescheduling)

    def test_default_currency_setting(self):
        """The Default currency is USD."""
        self.assertEqual(self.service.currency, 'USD')

        # Change the currency to EUR
        s = deepcopy(self.service)
        s.currency = 'EUR'
        self.assertEqual(s.currency, 'EUR')


class ServiceModelNegativeTestCase(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_exceeding_service_name_length(self):
        """Test that the max_length of the name field is 100 characters."""
        s = deepcopy(self.service)
        s.name = ("Gate Diagnostics and Calibration for Intergalactic Travel through the Quantum Bridge Device "
                  "Portal - Series SG-1")  # Exceeding the max_length (112 characters)
        with self.assertRaises(ValidationError):
            s.full_clean()

    def test_negative_price_and_down_payment_values(self):
        """Test that the price and down_payment fields cannot be negative."""
        s = deepcopy(self.service)
        s.price = -100
        with self.assertRaises(ValidationError):
            s.full_clean()

        s = deepcopy(self.service)
        s.down_payment = -100
        with self.assertRaises(ValidationError):
            s.full_clean()

    def test_invalid_currency_code_length(self):
        """A service cannot be created with a currency of less or more than three characters."""
        s = deepcopy(self.service)
        s.currency = "US"

        with self.assertRaises(ValidationError):
            s.full_clean()

        s.currency = "DOLLAR"
        with self.assertRaises(ValidationError):
            s.full_clean()

    def test_zero_or_negative_duration_handling(self):
        """A service cannot be created with a duration being zero or negative."""
        service = Service(name="Zat'nik'tel Tune-Up", duration=timedelta(0), price=100, description="Tune-up the Zat")
        self.assertRaises(ValidationError, service.full_clean)
        service = Service(name="Ancient's Archive Retrieval ", duration=timedelta(seconds=-1), price=50,
                          description="Retrieve the Ancient's Archive")
        self.assertRaises(ValidationError, service.full_clean)

    def test_creation_without_service_name(self):
        """A service cannot be created with no name."""
        with self.assertRaises(ValidationError):
            Service.objects.create(name="", duration=timedelta(hours=1), price=100).full_clean()



import datetime
import time

from django.utils import timezone

from appointment.models import PasswordResetToken
from appointment.tests.base.base_test import BaseTest


class PasswordResetTokenCreationTests(BaseTest):

    def setUp(self):
        super().setUp()
        self.user = self.create_user_(username='janet.fraiser', email='janet.fraiser@django-appointment.com',
                                      password='LovedCassandra', first_name='Janet')
        self.expired_time = timezone.now() - datetime.timedelta(minutes=5)
        self.token = PasswordResetToken.create_token(user=self.user)

    def tearDown(self):
        super().tearDown()
        self.user.delete()
        self.token.delete()

    def test_default_attributes_on_creation(self):
        self.assertIsNotNone(self.token)
        self.assertFalse(self.token.is_expired)
        self.assertFalse(self.token.is_verified)

    def test_str_representation(self):
        """Test the string representation of the token."""
        expected_str = (f"Password reset token for {self.user} "
                        f"[{self.token.token} status: {self.token.status} expires at {self.token.expires_at}]")
        self.assertEqual(str(self.token), expected_str)


class PasswordResetTokenPropertiesTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.user = self.create_user_(username='janet.fraiser', email='janet.fraiser@django-appointment.com',
                                      password='LovedCassandra', first_name='Janet')
        self.expired_time = timezone.now() - datetime.timedelta(minutes=5)
        self.token = PasswordResetToken.create_token(user=self.user)

    def tearDown(self):
        super().tearDown()
        self.user.delete()
        self.token.delete()

    def test_is_verified_property(self):
        """Test the is_verified property to check if the token status is correctly identified as verified."""
        token = PasswordResetToken.create_token(self.user)
        self.assertFalse(token.is_verified, "Newly created token should not be verified.")
        token.mark_as_verified()
        self.assertTrue(token.is_verified, "Token should be marked as verified after calling mark_as_verified.")

    def test_is_active_property(self):
        """Test the is_active property to check if the token status is correctly identified as active."""
        token = PasswordResetToken.create_token(self.user)
        self.assertTrue(token.is_active, "Newly created token should be active.")
        token.mark_as_verified()
        token.refresh_from_db()
        self.assertFalse(token.is_active, "Token should not be active after being verified.")

        # Invalidate the token and check is_active property
        token.status = PasswordResetToken.TokenStatus.INVALIDATED
        token.save()
        self.assertFalse(token.is_active, "Token should not be active after being invalidated.")

    def test_is_invalidated_property(self):
        """Test the is_invalidated property to check if the token status is correctly identified as invalidated."""
        token = PasswordResetToken.create_token(self.user)
        self.assertFalse(token.is_invalidated, "Newly created token should not be invalidated.")

        # Invalidate the token and check is_invalidated property
        token.status = PasswordResetToken.TokenStatus.INVALIDATED
        token.save()
        self.assertTrue(token.is_invalidated, "Token should be marked as invalidated after status change.")

    def test_token_expiration(self):
        """Test that a token is considered expired after the expiration time."""
        token = PasswordResetToken.create_token(user=self.user, expiration_minutes=-1)  # Token already expired
        self.assertTrue(token.is_expired)


class PasswordResetTokenVerificationTests(BaseTest):
    def setUp(self):
        super().setUp()
        self.user = self.create_user_(username='janet.fraiser', email='janet.fraiser@django-appointment.com',
                                      password='LovedCassandra', first_name='Janet')
        self.expired_time = timezone.now() - datetime.timedelta(minutes=5)

    def test_verify_token_success(self):
        """Test successful token verification."""
        token = PasswordResetToken.create_token(user=self.user)
        verified_token = PasswordResetToken.verify_token(user=self.user, token=token.token)
        self.assertIsNotNone(verified_token)

    def test_verify_token_failure_expired(self):
        """Test token verification fails if the token has expired."""
        token = PasswordResetToken.create_token(user=self.user, expiration_minutes=-1)  # Token already expired
        verified_token = PasswordResetToken.verify_token(user=self.user, token=token.token)

        self.assertIsNone(verified_token, "Expired token should not verify")

    def test_verify_token_failure_wrong_user(self):
        """Test token verification fails if the token does not belong to the given user."""
        another_user = self.create_user_(username='another_user', email='another@example.com',
                                         password='test_pass456')
        token = PasswordResetToken.create_token(user=self.user)
        verified_token = PasswordResetToken.verify_token(user=another_user, token=token.token)
        self.assertIsNone(verified_token)

    def test_verify_token_failure_already_verified(self):
        """Test token verification fails if the token has already been verified."""
        token = PasswordResetToken.create_token(user=self.user)
        token.mark_as_verified()
        verified_token = PasswordResetToken.verify_token(user=self.user, token=token.token)
        self.assertIsNone(verified_token)

    def test_verify_token_invalid_token(self):
        """Test token verification fails if the token does not exist."""
        PasswordResetToken.create_token(user=self.user)
        invalid_token_uuid = "12345678-1234-1234-1234-123456789012"  # An invalid token UUID
        verified_token = PasswordResetToken.verify_token(user=self.user, token=invalid_token_uuid)
        self.assertIsNone(verified_token)

    def test_token_expiration_boundary(self):
        """Test token verification at the exact moment of expiration."""
        token = PasswordResetToken.create_token(user=self.user, expiration_minutes=0)  # Token expires now
        # Assuming there might be a very slight delay before verification, we wait a second
        time.sleep(1)
        verified_token = PasswordResetToken.verify_token(user=self.user, token=token.token)
        self.assertIsNone(verified_token)

    def test_create_multiple_tokens_for_user(self):
        """Test that multiple tokens can be created for a single user and only the latest is valid."""
        old_token = PasswordResetToken.create_token(user=self.user)
        new_token = PasswordResetToken.create_token(user=self.user)

        old_verified = PasswordResetToken.verify_token(user=self.user, token=old_token.token)
        new_verified = PasswordResetToken.verify_token(user=self.user, token=new_token.token)

        self.assertIsNone(old_verified, "Old token should not be valid after creating a new one")
        self.assertIsNotNone(new_verified, "New token should be valid")

    def test_token_verification_resets_after_expiration(self):
        """Test that an expired token cannot be verified after its expiration, even if marked as verified."""
        token = PasswordResetToken.create_token(user=self.user, expiration_minutes=-1)  # Already expired
        token.mark_as_verified()

        verified_token = PasswordResetToken.verify_token(user=self.user, token=token.token)
        self.assertIsNone(verified_token, "Expired token should not verify, even if marked as verified")

    def test_verify_token_invalidated(self):
        """Test token verification fails if the token has been invalidated."""
        token = PasswordResetToken.create_token(self.user)
        # Invalidate the token by creating a new one
        PasswordResetToken.create_token(self.user)
        verified_token = PasswordResetToken.verify_token(self.user, token.token)
        self.assertIsNone(verified_token)

    def test_expired_token_verification(self):
        """Test that an expired token cannot be verified."""
        token = PasswordResetToken.objects.create(user=self.user, expires_at=self.expired_time,
                                                  status=PasswordResetToken.TokenStatus.ACTIVE)
        self.assertTrue(token.is_expired)
        verified_token = PasswordResetToken.verify_token(self.user, token.token)
        self.assertIsNone(verified_token, "Expired token should not verify")

    def test_token_verification_after_user_deletion(self):
        """Test that a token cannot be verified after the associated user is deleted."""
        token = PasswordResetToken.create_token(self.user)
        self.user.delete()
        verified_token = PasswordResetToken.verify_token(None, token.token)
        self.assertIsNone(verified_token, "Token should not verify after user deletion")


class PasswordResetTokenTests(BaseTest):
    def setUp(self):
        super().setUp()
        self.user = self.create_user_(username='janet.fraiser', email='janet.fraiser@django-appointment.com',
                                      password='LovedCassandra', first_name='Janet')
        self.expired_time = timezone.now() - datetime.timedelta(minutes=5)

    def tearDown(self):
        super().tearDown()
        PasswordResetToken.objects.all().delete()
        self.user.delete()

    def test_mark_as_verified(self):
        """Test marking a token as verified."""
        token = PasswordResetToken.create_token(user=self.user)
        self.assertFalse(token.is_verified)
        token.mark_as_verified()
        token.refresh_from_db()  # Refresh the token object from the database
        self.assertTrue(token.is_verified)

    def test_mark_as_verified_is_idempotent(self):
        """Test that marking a token as verified multiple times has no adverse effect."""
        token = PasswordResetToken.create_token(user=self.user)
        token.mark_as_verified()
        first_verification_time = token.updated_at

        time.sleep(1)  # Ensure time has passed
        token.mark_as_verified()
        token.refresh_from_db()

        self.assertTrue(token.is_verified)
        self.assertEqual(first_verification_time, token.updated_at,
                         "Token verification time should not update on subsequent calls")

    def test_deleting_user_cascades_to_tokens(self):
        """Test that deleting a user deletes associated password reset tokens."""

        apophis = self.create_user_(username='apophis.false_god', email='apophis.false_god@django-appointment.com',
                                    password='LovedSayingSholva', first_name='Apophis')
        token = PasswordResetToken.create_token(user=apophis)
        apophis.delete()

        with self.assertRaises(PasswordResetToken.DoesNotExist):
            PasswordResetToken.objects.get(pk=token.pk)



# admin.py
# Path: appointment/admin.py

"""
Author: Adams Pierre David
Since: 1.0.0
"""

from django import forms
from django.contrib import admin

from .models import (
    Appointment, AppointmentRequest, AppointmentRescheduleHistory, Config, DayOff, EmailVerificationCode,
    PasswordResetToken, Service, StaffMember, WorkingHours
)


@admin.register(Service)
class ServiceAdmin(admin.ModelAdmin):
    list_display = ('name', 'duration', 'price', 'created_at', 'updated_at',)
    search_fields = ('name',)
    list_filter = ('duration',)


@admin.register(AppointmentRequest)
class AppointmentRequestAdmin(admin.ModelAdmin):
    list_display = ('date', 'start_time', 'end_time', 'service', 'created_at', 'updated_at',)
    search_fields = ('date', 'service__name',)
    list_filter = ('date', 'service',)


@admin.register(Appointment)
class AppointmentAdmin(admin.ModelAdmin):
    list_display = ('client', 'appointment_request', 'created_at', 'updated_at',)
    search_fields = ('appointment_request__service__name',)
    list_filter = ('client', 'appointment_request__service',)


@admin.register(EmailVerificationCode)
class EmailVerificationCodeAdmin(admin.ModelAdmin):
    list_display = ('user', 'code')


@admin.register(Config)
class ConfigAdmin(admin.ModelAdmin):
    list_display = (
        'slot_duration', 'lead_time', 'finish_time', 'appointment_buffer_time', 'website_name', 'app_offered_by_label')


# Define a custom ModelForm for StaffMember
class StaffMemberForm(forms.ModelForm):
    class Meta:
        model = StaffMember
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@admin.register(StaffMember)
class StaffMemberAdmin(admin.ModelAdmin):
    form = StaffMemberForm
    list_display = (
        'get_staff_member_name', 'get_slot_duration', 'lead_time', 'finish_time', 'work_on_saturday', 'work_on_sunday')
    search_fields = ('user__email', 'user__first_name', 'user__last_name')
    list_filter = ('work_on_saturday', 'work_on_sunday', 'lead_time', 'finish_time')


@admin.register(DayOff)
class DayOffAdmin(admin.ModelAdmin):
    list_display = ('staff_member', 'start_date', 'end_date', 'description')
    search_fields = ('description',)
    list_filter = ('start_date', 'end_date')


@admin.register(WorkingHours)
class WorkingHoursAdmin(admin.ModelAdmin):
    list_display = ('staff_member', 'day_of_week', 'start_time', 'end_time')
    search_fields = ('day_of_week',)
    list_filter = ('day_of_week', 'start_time', 'end_time')


@admin.register(AppointmentRescheduleHistory)
class AppointmentRescheduleHistoryAdmin(admin.ModelAdmin):
    list_display = (
        'appointment_request', 'date', 'start_time',
        'end_time', 'staff_member', 'reason_for_rescheduling', 'created_at'
    )
    search_fields = (
        'appointment_request__id_request', 'staff_member__user__first_name',
        'staff_member__user__last_name', 'reason_for_rescheduling'
    )
    list_filter = ('appointment_request__service', 'date', 'created_at')
    date_hierarchy = 'created_at'
    ordering = ('-created_at',)


@admin.register(PasswordResetToken)
class PasswordResetTokenAdmin(admin.ModelAdmin):
    list_display = ('user', 'token', 'expires_at', 'status')
    search_fields = ('user__email', 'token')
    list_filter = ('status', 'expires_at')
    date_hierarchy = 'expires_at'
    ordering = ('-expires_at',)



# services.py
# Path: appointment/services.py

"""
Author: Adams Pierre David
Since: 2.0.0
"""

import datetime

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import render
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext as _, gettext_lazy as _

from appointment.forms import PersonalInformationForm, ServiceForm, StaffDaysOffForm, StaffWorkingHoursForm
from appointment.messages_ import appt_updated_successfully
from appointment.settings import APPOINTMENT_PAYMENT_URL
from appointment.utils.date_time import (
    convert_12_hour_time_to_24_hour_time, convert_str_to_date, convert_str_to_time, get_ar_end_time)
from appointment.utils.db_helpers import (
    Appointment, AppointmentRequest, EmailVerificationCode, Service, StaffMember, WorkingHours, calculate_slots,
    calculate_staff_slots, check_day_off_for_staff, create_and_save_appointment, create_new_user,
    day_off_exists_for_date_range, exclude_booked_slots, exclude_pending_reschedules, get_all_appointments,
    get_all_staff_members,
    get_appointment_by_id, get_appointments_for_date_and_time, get_staff_member_appointment_list,
    get_staff_member_from_user_id_or_logged_in, get_times_from_config, get_user_by_email,
    get_weekday_num_from_date, get_working_hours_for_staff_and_day, parse_name, update_appointment_reminder,
    working_hours_exist)
from appointment.utils.email_ops import send_reset_link_to_staff_member
from appointment.utils.error_codes import ErrorCode
from appointment.utils.json_context import convert_appointment_to_json, get_generic_context, json_response
from appointment.utils.permissions import check_entity_ownership
from appointment.utils.session import handle_email_change


def fetch_user_appointments(user):
    """Fetch the appointments for a given user.

    :param user: The user instance.
    :return: A list of appointments.
    """
    if user.is_superuser:
        return get_all_appointments()
    try:
        staff_member_instance = user.staffmember
        return get_staff_member_appointment_list(staff_member_instance)
    except ObjectDoesNotExist:
        if user.is_staff:
            return []

    raise ValueError("User is not a staff member or a superuser")


def prepare_appointment_display_data(user, appointment_id):
    """Prepare the data for the appointment details page.

    :param user: The user instance.
    :param appointment_id: The appointment id.
    :return: A tuple containing the appointment instance, page title, error message, and status code.
    """
    appointment = get_appointment_by_id(appointment_id)

    # If the appointment doesn't exist
    if not appointment:
        return None, None, _("Appointment does not exist."), 404

    # Check if the user is authorized to view the appointment
    if not check_entity_ownership(user, appointment):
        return None, None, _("You are not authorized to view this appointment."), 403

    # Prepare the data for display
    page_title = _("Appointment details") + _(": {client_name}").format(client_name=appointment.get_client_name())
    if user.is_superuser:
        page_title += f' (by: {appointment.get_staff_member_name()})'

    return appointment, page_title, None, 200


def prepare_user_profile_data(user, staff_user_id):
    """Prepare the data for the user profile page.

    :param user: The user instance.
    :param staff_user_id: The staff user id.
    :return: A dictionary containing the data for the user profile page.
    """
    if user.is_superuser and staff_user_id is None:
        staff_members = get_all_staff_members()
        btn_staff_me = "Staff me"
        btn_staff_me_link = reverse('appointment:make_superuser_staff_member')
        if StaffMember.objects.filter(user=user).exists():
            btn_staff_me = "Remove me"
            btn_staff_me_link = reverse('appointment:remove_superuser_staff_member')
        data = {
            'error': False,
            'template': 'administration/staff_list.html',
            'extra_context': {
                'staff_members': staff_members,
                'btn_staff_me': btn_staff_me,
                'btn_staff_me_link': btn_staff_me_link
            }
        }
        return data

    if staff_user_id and staff_user_id != user.pk and not user.is_superuser:
        return {
            'error': True,
            'extra_context': {'message': _("You can only view your own profile"),
                              'back_url': reverse('appointment:user_profile')},
            'status_code': 403
        }
    staff_member = get_staff_member_from_user_id_or_logged_in(user, staff_user_id)

    if not staff_member:
        return {
            'error': True,
            'extra_context': {'message': _("Not authorized."), 'back_url': reverse('appointment:user_profile')},
            'status_code': 403
        }

    bt_help = StaffMember._meta.get_field('appointment_buffer_time')
    bt_help_text = bt_help.help_text

    sd_help = StaffMember._meta.get_field('slot_duration')
    sd_help_text = sd_help.help_text
    if user.is_superuser:
        service_msg = _("Here you can add/remove services offered by this staff member by modifying this section.")
    else:
        service_msg = _("Here you can add/remove services offered by you by modifying this section.")
    return {
        'error': False,
        'template': 'administration/user_profile.html',
        'extra_context': {
            'superuser': user if user.is_superuser else None,
            'user': staff_member.user if staff_member else user,
            'staff_member': staff_member,
            'days_off': staff_member.get_days_off().order_by('start_date') if staff_member else [],
            'working_hours': staff_member.get_working_hours() if staff_member else [],
            'services_offered': staff_member.get_services_offered() if staff_member else [],
            'staff_member_not_found': not bool(staff_member),
            'buffer_time_help_text': bt_help_text,
            'slot_duration_help_text': sd_help_text,
            'service_msg': service_msg,
        }
    }


###############################################################
# handler for adding, updating, and deleting day off and working hours

def handle_entity_management_request(request, staff_member, entity_type, instance=None, staff_user_id=None,
                                     instance_id=None, add=True):
    """Handle the request to add, update, or delete a day off or working hours.

    :param request: The request object.
    :param staff_member: The staff member instance.
    :param entity_type: The type of entity to add or update, either 'day_off' or 'working_hours'.
    :param instance: The instance of the entity to update.
    :param staff_user_id: The staff user id.
    :param instance_id: The id of the instance to update.
    :param add: If True, add a new entity, otherwise, update an existing one.
    :return: A JsonResponse instance.
    """
    if not staff_member:
        return json_response("Not authorized", status=403, success=False,
                             error_code=ErrorCode.NOT_AUTHORIZED)

    button_text = _('Update') if instance else _('Add')
    if entity_type == 'day_off':
        form = StaffDaysOffForm(instance=instance)
        context = get_working_hours_and_days_off_context(request, button_text, 'day_off_form', form)
        template = 'administration/manage_day_off.html'
    else:
        form = StaffWorkingHoursForm(instance=instance)
        context = get_working_hours_and_days_off_context(request, button_text, 'working_hours_form', form,
                                                         staff_user_id, instance,
                                                         instance_id)
        template = 'administration/manage_working_hours.html'

    if request.method == 'POST' and entity_type == 'day_off':
        day_off_form = StaffDaysOffForm(request.POST, instance=instance)
        start_date = request.POST.get('start_date')
        end_date = request.POST.get('end_date')

        if day_off_exists_for_date_range(staff_member, start_date, end_date, getattr(instance, 'id', None)):
            return json_response(_("Days off for this date range already exist."), status=400, success=False,
                                 error_code=ErrorCode.DAY_OFF_CONFLICT)

        return handle_day_off_form(day_off_form, staff_member)
    elif request.method == 'POST' and entity_type == 'working_hours':
        day_of_week = request.POST.get('day_of_week')
        start_time = request.POST.get('start_time')
        end_time = request.POST.get('end_time')

        return handle_working_hours_form(staff_member, day_of_week, start_time, end_time, add, instance_id)

    return render(request, template, context, status=200)


def handle_day_off_form(day_off_form, staff_member):
    """Handle the day off form.

    :param day_off_form: The day off form instance.
    :param staff_member: The staff member instance.
    :return: A JsonResponse instance.
    """
    if day_off_form.is_valid():
        day_off = day_off_form.save(commit=False)
        day_off.staff_member = staff_member
        day_off.save()
        redirect_url = reverse('appointment:user_profile',
                               kwargs={'staff_user_id': staff_member.user_id}) if staff_member else reverse(
            'appointment:user_profile')
        return json_response(_("Day off saved successfully."), custom_data={'redirect_url': redirect_url})
    else:
        message = "Invalid data:"
        message += get_error_message_in_form(form=day_off_form)
        return json_response(message, status=400, success=False, error_code=ErrorCode.INVALID_DATA)


def handle_working_hours_form(staff_member, day_of_week, start_time, end_time, add, wh_id=None):
    """Handle the working hours form.

    :param staff_member: The staff member instance.
    :param day_of_week: The day of the week.
    :param start_time: The start time.
    :param end_time: The end time.
    :param add: If True, add a new working hours instance. Otherwise, update an existing one.
    :param wh_id: The working hours' id.
    :return: A JsonResponse instance.
    """
    # Validate inputs
    if not (staff_member and day_of_week and start_time and end_time):
        return json_response(_("Invalid data."), status=400, success=False, error_code=ErrorCode.INVALID_DATA)

    # Convert start time and end time to 24-hour format
    start_time = convert_12_hour_time_to_24_hour_time(start_time)
    end_time = convert_12_hour_time_to_24_hour_time(end_time)

    # Ensure start time is before end time
    if start_time >= end_time:
        return json_response(_("Start time must be before end time."), status=400, success=False,
                             error_code=ErrorCode.INVALID_DATA)

    if add:
        # Create new working hours
        if working_hours_exist(day_of_week=day_of_week, staff_member=staff_member):
            return json_response(_("Working hours already exist for this day."), status=400, success=False,
                                 error_code=ErrorCode.WORKING_HOURS_CONFLICT)
        wk = WorkingHours(staff_member=staff_member, day_of_week=day_of_week, start_time=start_time, end_time=end_time)
    else:
        # Ensure working_hours_id is provided
        if not wh_id:
            return json_response(_("Invalid or no working_hours_id provided."), status=400, success=False,
                                 error_code=ErrorCode.INVALID_DATA)

        # Get the working hours instance to update
        try:
            wk = WorkingHours.objects.get(pk=wh_id)
            wk.day_of_week = day_of_week
            wk.start_time = start_time
            wk.end_time = end_time
        except WorkingHours.DoesNotExist:
            return json_response(_("Working hours does not exist."), status=400, success=False,
                                 error_code=ErrorCode.WORKING_HOURS_NOT_FOUND)

        # Save working hours
    wk.save()

    # Return success with redirect URL
    redirect_url = reverse('appointment:user_profile', kwargs={'staff_user_id': staff_member.user.id}) \
        if staff_member.user.id else reverse('appointment:user_profile')
    return json_response(_("Working hours saved successfully."), custom_data={'redirect_url': redirect_url})


def get_working_hours_and_days_off_context(request, btn_txt, form_name, form, user_id=None, instance=None, wh_id=None):
    """Get the context for the working hours and days off forms.

    :param request: The request object.
    :param btn_txt: The text to display on the submit button.
    :param form_name: The name of the form which depends on if it's a working hours or days off form.
    :param form: The form instance itself.
    :param user_id: The staff user id.
    :param instance: The working hour form instance.
    :param wh_id: The working hour id.
    :return: A dictionary containing the context.
    """
    context = get_generic_context(request)
    context.update({
        'button_text': btn_txt,
        form_name: form,
    })
    if user_id:
        context.update({
            'staff_user_id': user_id,
        })
    if instance:
        context.update({
            'working_hours_instance': instance,
        })
    if wh_id:
        context.update({
            'working_hours_id': wh_id,
        })
    return context


def save_appointment(appt, client_name, client_email, start_time, phone_number, client_address, service_id, request,
                     staff_member_id=None, want_reminder=False, additional_info=None):
    """Save an appointment's details.
    :return: The modified appointment.
    """
    service = Service.objects.get(id=service_id)
    if staff_member_id:
        staff_member = StaffMember.objects.get(id=staff_member_id)
        if not staff_member.get_service_is_offered(service_id):
            return None
    # Modify and save client details
    first_name, last_name = parse_name(client_name)
    client = appt.client
    client.first_name = first_name
    client.last_name = last_name
    client.email = client_email
    client.save()
    # convert start time to a time object if it is a string
    if isinstance(start_time, str):
        start_time = convert_str_to_time(start_time)
    # calculate end time from start time and service duration
    end_time = get_ar_end_time(start_time, service.duration)

    # Modify and save appointment request details
    appt_request = appt.appointment_request

    # Update reminder here
    update_appointment_reminder(appointment=appt, new_date=appt_request.date, new_start_time=start_time,
                                want_reminder=want_reminder, request=request)

    appt_request.service = service
    appt_request.start_time = start_time
    appt_request.end_time = end_time
    appt_request.staff_member = staff_member
    appt_request.save()

    # Modify and save appointment details
    appt.phone = phone_number
    appt.address = client_address
    appt.want_reminder = want_reminder
    appt.additional_info = additional_info
    appt.save()
    return appt


def save_appt_date_time(appt_start_time, appt_date, appt_id, request):
    """Save the date and time of an appointment request.

    :param appt_start_time: The start time of the appointment request.
    :param appt_date: The date of the appointment request.
    :param appt_id: The ID of the appointment to modify.
    :param request: The request object.
    :return: The modified appointment.
    """
    appt = Appointment.objects.get(id=appt_id)
    service = appt.get_service()

    # Convert start time to a time object if it is a string
    if isinstance(appt_start_time, str):
        time_format = "%H:%M:%S.%fZ"
        appt_start_time_obj = datetime.datetime.strptime(appt_start_time, time_format).time()
    else:
        appt_start_time_obj = appt_start_time

    # Calculate end time from start time and service duration
    end_time_obj = get_ar_end_time(appt_start_time_obj, service.duration)

    # Convert the appt_date from string to a date object if it's a string
    if isinstance(appt_date, str):
        appt_date_obj = datetime.datetime.strptime(appt_date, "%Y-%m-%d").date()
    else:
        appt_date_obj = appt_date

    # Update reminder here
    update_appointment_reminder(appointment=appt, new_date=appt_date_obj, new_start_time=appt_start_time_obj,
                                request=request)

    # Modify and save appointment request details
    appt_request = appt.appointment_request
    appt_request.date = appt_date_obj
    appt_request.start_time = appt_start_time_obj
    appt_request.end_time = end_time_obj
    appt_request.save()
    appt.save()

    return appt


def get_available_slots(date, appointments):
    """Calculate the available time slots for a given date and a list of appointments.

    :param date: The date for which to calculate the available slot
    :param appointments: A list of Appointment objects
    :return: A list of available time slots as strings in the format '%I:%M %p' like ['10:00 AM', '10:30 AM']
    """

    start_time, end_time, slot_duration, buff_time = get_times_from_config(date)
    now = timezone.now()
    buffer_time = now + buff_time if date == now.date() else now
    slots = calculate_slots(start_time, end_time, buffer_time, slot_duration)
    slots = exclude_booked_slots(appointments, slots, slot_duration)
    return [slot.strftime('%I:%M %p') for slot in slots]


def get_available_slots_for_staff(date, staff_member):
    """Calculate the available time slots for a given date and a staff member.

    :param date: The date for which to calculate the available slots
    :param staff_member: The staff member for which to calculate the available slots
    :return: A list of available time slots as strings in the format '%I:%M %p' like ['10:00 AM', '10:30 AM']
    """
    # Check if the provided date is a day off for the staff member
    days_off_exist = check_day_off_for_staff(staff_member=staff_member, date=date)
    if days_off_exist:
        return []

    # Check if the staff member works on the provided date
    day_of_week = get_weekday_num_from_date()  # Python's weekday starts from Monday (0) to Sunday (6)
    working_hours_dict = get_working_hours_for_staff_and_day(staff_member, day_of_week)
    if not working_hours_dict:
        return []

    slot_duration = datetime.timedelta(minutes=staff_member.get_slot_duration())
    slots = calculate_staff_slots(date, staff_member)
    slots = exclude_pending_reschedules(slots, staff_member, date)
    appointments = get_appointments_for_date_and_time(date, working_hours_dict['start_time'],
                                                      working_hours_dict['end_time'], staff_member)
    return exclude_booked_slots(appointments, slots, slot_duration)


def get_finish_button_text(service) -> str:
    """
    Check if a service is free.

    :param service: Service, the service to check
    :return: bool, True if the service is free, False otherwise
    """
    if service.is_a_paid_service() and APPOINTMENT_PAYMENT_URL:
        return _("Pay Now")
    return _("Finish")


def get_appointments_and_slots(date_, service=None):
    """
    Get appointments and available slots for a given date and service.

    If a service is provided, the function retrieves appointments for that service on the given date.
    Otherwise, it retrieves all appointments for the given date.

    :param date_: datetime.date, the date for which to retrieve appointments and available slots
    :param service: Service, the service for which to retrieve appointments
    :return: tuple, a tuple containing two elements:
        - A queryset of appointments for the given date and service (if provided).
        - A list of available time slots on the given date, excluding booked appointments.
    """
    if service:
        appointments = Appointment.objects.filter(appointment_request__service=service,
                                                  appointment_request__date=date_)
    else:
        appointments = Appointment.objects.filter(appointment_request__date=date_)
    available_slots = get_available_slots(date_, appointments)
    return appointments, available_slots


def get_error_message_in_form(form):
    """
    Get the error message in a form.

    :param form: Form, the form to check
    :return: str, the error message
    """
    error_messages = []
    for field, errors in form.errors.items():
        error_messages.append(f"{field}: {','.join(errors)}")
    if len(error_messages) == 3:
        return _("Empty fields are not allowed.")
    return " ".join(error_messages)


def update_personal_info_service(staff_user_id, post_data, current_user):
    try:
        user = get_user_model().objects.get(pk=staff_user_id) if staff_user_id else current_user
    except get_user_model().DoesNotExist:
        return None, False, _("User not found.")

    form = PersonalInformationForm(post_data, user=user)
    if form.is_valid():
        user.first_name = form.cleaned_data['first_name']
        user.last_name = form.cleaned_data['last_name']
        if form.cleaned_data['email'] != user.email:
            handle_email_change(current_user, user, form.cleaned_data['email'])
        user.save()
        return user, True, None
    else:
        return None, False, get_error_message_in_form(form=form)


def email_change_verification_service(code, email, old_email):
    user = get_user_by_email(old_email)
    code_obj = EmailVerificationCode.objects.filter(code=code).first()

    if user and code_obj and code_obj.check_code(code):
        user.email = email
        user.save()
        code_obj.delete()
        return True
    return False


def create_staff_member_service(post_data, request):
    form = PersonalInformationForm(post_data)
    if form.is_valid():
        first_name = form.cleaned_data['first_name']
        last_name = form.cleaned_data['last_name']
        email = form.cleaned_data["email"]

        if get_user_by_email(email):
            return None, False, _("A user with this email already exists.")

        user_data = {
            'first_name': first_name,
            'last_name': last_name,
            'email': email,
        }
        user = create_new_user(client_data=user_data)
        StaffMember.objects.create(user=user)
        if not user.is_superuser:
            user.is_staff = True
            user.save()
        send_reset_link_to_staff_member(user, request, user.email)
        return user, True, None
    else:
        return None, False, get_error_message_in_form(form=form)


def handle_service_management_request(post_data, files_data=None, service_id=None):
    try:
        if service_id:
            service = Service.objects.get(pk=service_id)
            form = ServiceForm(post_data, files_data, instance=service)
        else:
            form = ServiceForm(post_data, files_data)

        if form.is_valid():
            form.save()
            return form.instance, True, _("Service saved successfully.")
        else:
            return None, False, get_error_message_in_form(form=form)
    except Exception as e:
        return None, False, str(e)


def create_new_appointment(data, request):
    service = Service.objects.get(id=data.get("service_id"))
    staff_id = data.get("staff_member")
    if staff_id:
        staff_member = StaffMember.objects.get(id=staff_id)
    else:
        staff_member = StaffMember.objects.get(user=request.user)

    # Convert date and start time strings to datetime objects
    date = convert_str_to_date(data.get("date"))
    start_time = convert_str_to_time(data.get("start_time"))
    start_datetime = datetime.datetime.combine(date, start_time)

    appointment_request = AppointmentRequest(
        date=start_datetime.date(),
        start_time=start_datetime.time(),
        end_time=(start_datetime + service.duration).time(),
        service=service,
        staff_member=staff_member,
    )
    appointment_request.full_clean()  # Validates the model
    appointment_request.save()

    # Prepare client data
    email = data.get("client_email", "")
    name_parts = data.get("client_name", "").split()
    first_name = name_parts[0] if name_parts else ""
    last_name = name_parts[-1] if len(name_parts) > 1 else ""  # Use an empty string if no last name

    client_data = {
        'email': email,
        'first_name': first_name,
        'last_name': last_name,
    }

    # Use your custom user creation logic
    user = get_user_by_email(email)
    if not user:
        create_new_user(client_data)

    # Create and save the new appointment
    appointment_data = {
        'phone': data.get("client_phone", ""),
        'address': data.get("client_address", ""),
        'want_reminder': data.get("want_reminder") == 'true',
        'additional_info': data.get("additional_info", ""),
        'paid': False
    }
    appointment = create_and_save_appointment(appointment_request, client_data, appointment_data, request)
    appointment_list = convert_appointment_to_json(request, [appointment])

    return json_response("Appointment created successfully.", custom_data={'appt': appointment_list})


def update_existing_appointment(data, request):
    try:
        appt = Appointment.objects.get(id=data.get("appointment_id"))
        staff_id = data.get("staff_member")
        want_reminder = data.get("want_reminder") == 'true'
        appt = save_appointment(
            appt,
            client_name=data.get("client_name"),
            client_email=data.get("client_email"),
            start_time=data.get("start_time"),
            phone_number=data.get("client_phone"),
            client_address=data.get("client_address"),
            service_id=data.get("service_id"),
            want_reminder=want_reminder,
            additional_info=data.get("additional_info"),
            staff_member_id=staff_id,
            request=request,
        )
        if not appt:
            return json_response("Service not offered by staff member.", status=400, success=False,
                                 error_code=ErrorCode.SERVICE_NOT_FOUND)
        appointments_json = convert_appointment_to_json(request, [appt])[0]
        return json_response(appt_updated_successfully, custom_data={'appt': appointments_json})
    except Appointment.DoesNotExist:
        return json_response("Appointment does not exist.", status=404, success=False,
                             error_code=ErrorCode.APPOINTMENT_NOT_FOUND)
    except Service.DoesNotExist:
        return json_response("Service does not exist.", status=404, success=False,
                             error_code=ErrorCode.SERVICE_NOT_FOUND)
    except Exception as e:
        return json_response(str(e.args[0]), status=400, success=False)



# forms.py
# Path: appointment/forms.py

"""
Author: Adams Pierre David
Since: 1.0.0
"""

from django import forms
from django.utils.translation import gettext as _
from phonenumber_field.formfields import SplitPhoneNumberField

from .models import (
    Appointment, AppointmentRequest, AppointmentRescheduleHistory, DayOff, Service, StaffMember,
    WorkingHours
)
from .utils.db_helpers import get_user_model
from .utils.validators import not_in_the_past


class SlotForm(forms.Form):
    selected_date = forms.DateField(validators=[not_in_the_past])
    staff_member = forms.ModelChoiceField(StaffMember.objects.all(),
                                          error_messages={'invalid_choice': 'Staff member does not exist'})


class AppointmentRequestForm(forms.ModelForm):
    class Meta:
        model = AppointmentRequest
        fields = ('date', 'start_time', 'end_time', 'service', 'staff_member')


class ReschedulingForm(forms.ModelForm):
    class Meta:
        model = AppointmentRescheduleHistory
        fields = ['reason_for_rescheduling']
        widgets = {
            'reason_for_rescheduling': forms.Textarea(attrs={'rows': 4, 'placeholder': 'Reason for rescheduling...'}),
        }


class AppointmentForm(forms.ModelForm):
    phone = SplitPhoneNumberField()

    class Meta:
        model = Appointment
        fields = ('phone', 'want_reminder', 'address', 'additional_info')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['phone'].widget.attrs.update(
                {
                    'placeholder': '1234567890'
                })
        self.fields['additional_info'].widget.attrs.update(
                {
                    'rows': 2,
                    'class': 'form-control',
                })
        self.fields['address'].widget.attrs.update(
                {
                    'rows': 2,
                    'class': 'form-control',
                    'placeholder': '1234 Main St, City, State, Zip Code',
                    'required': 'true'
                })
        self.fields['additional_info'].widget.attrs.update(
                {
                    'class': 'form-control',
                    'placeholder': 'I would like to be contacted by phone.'
                })


class ClientDataForm(forms.Form):
    name = forms.CharField(max_length=50, widget=forms.TextInput(attrs={'class': 'form-control'}))
    email = forms.EmailField(widget=forms.EmailInput(attrs={'class': 'form-control'}))


class PersonalInformationForm(forms.Form):
    # first_name, last_name, email
    first_name = forms.CharField(max_length=50, widget=forms.TextInput(attrs={'class': 'form-control'}))
    last_name = forms.CharField(max_length=50, widget=forms.TextInput(attrs={'class': 'form-control'}))
    email = forms.EmailField(widget=forms.EmailInput(attrs={'class': 'form-control'}))

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop('user', None)  # pop the user from the kwargs
        super().__init__(*args, **kwargs)

    def clean_email(self):
        email = self.cleaned_data.get('email')
        if self.user:
            if self.user.email == email:
                return email
            queryset = get_user_model().objects.exclude(pk=self.user.pk)
        else:
            queryset = get_user_model().objects.all()

        if queryset.filter(email=email).exists():
            raise forms.ValidationError(_("This email is already taken."))

        return email


class StaffAppointmentInformationForm(forms.ModelForm):
    class Meta:
        model = StaffMember
        fields = ['services_offered', 'slot_duration', 'lead_time', 'finish_time',
                  'appointment_buffer_time', 'work_on_saturday', 'work_on_sunday']
        widgets = {
            'service_offered': forms.Select(attrs={'class': 'form-control'}),
            'slot_duration': forms.NumberInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 30, 60, 90, 120... (in minutes)')
            }),
            'lead_time': forms.TimeInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 08:00:00, 09:00:00... (24-hour format)')
            }),
            'finish_time': forms.TimeInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 17:00:00, 18:00:00... (24-hour format)')
            }),
            'appointment_buffer_time': forms.NumberInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 15, 30, 45, 60... (in minutes)')
            }),
            'work_on_saturday': forms.CheckboxInput(attrs={'class': 'form-check-input'}),
            'work_on_sunday': forms.CheckboxInput(attrs={'class': 'form-check-input'}),
        }


class StaffMemberForm(forms.ModelForm):
    class Meta:
        model = StaffMember
        fields = ['user', 'services_offered', 'slot_duration', 'lead_time', 'finish_time',
                  'appointment_buffer_time', 'work_on_saturday', 'work_on_sunday']
        widgets = {
            'user': forms.Select(attrs={'class': 'form-control'}),
            'service_offered': forms.Select(attrs={'class': 'form-control'}),
            'slot_duration': forms.NumberInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 30, 60, 90, 120... (in minutes)')
            }),
            'lead_time': forms.TimeInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 08:00:00, 09:00:00... (24-hour format)')
            }),
            'finish_time': forms.TimeInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 17:00:00, 18:00:00... (24-hour format)')
            }),
            'appointment_buffer_time': forms.NumberInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example value: 15, 30, 45, 60... (in minutes)')
            }),
            'work_on_saturday': forms.CheckboxInput(attrs={'class': 'form-check-input'}),
            'work_on_sunday': forms.CheckboxInput(attrs={'class': 'form-check-input'}),
        }

    def __init__(self, *args, **kwargs):
        super(StaffMemberForm, self).__init__(*args, **kwargs)
        # Exclude users who are already staff members
        existing_staff_user_ids = StaffMember.objects.values_list('user', flat=True)
        # Filter queryset for user field to include only superusers or users not already staff members
        self.fields['user'].queryset = get_user_model().objects.filter(
                is_superuser=True
        ).exclude(id__in=existing_staff_user_ids) | get_user_model().objects.exclude(
                id__in=existing_staff_user_ids
        )


class StaffDaysOffForm(forms.ModelForm):
    class Meta:
        model = DayOff
        fields = ['start_date', 'end_date', 'description']
        widgets = {
            'start_date': forms.DateInput(attrs={'class': 'datepicker'}),
            'end_date': forms.DateInput(attrs={'class': 'datepicker'}),
        }

    def clean(self):
        cleaned_data = super().clean()
        return cleaned_data


class StaffWorkingHoursForm(forms.ModelForm):
    class Meta:
        model = WorkingHours
        fields = ['day_of_week', 'start_time', 'end_time']


class ServiceForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(ServiceForm, self).__init__(*args, **kwargs)
        if self.instance and self.instance.pk:
            self.fields['background_color'].widget.attrs['value'] = self.instance.background_color

    class Meta:
        model = Service
        fields = ['name', 'description', 'duration', 'price', 'down_payment', 'image', 'currency', 'background_color']
        widgets = {
            'name': forms.TextInput(attrs={
                'class': 'form-control',
                'placeholder': _('Example: First Consultation')
            }),
            'description': forms.Textarea(attrs={
                'class': 'form-control',
                'placeholder': "Example: Overview of client's needs."
            }),
            'duration': forms.TextInput(attrs={
                'class': 'form-control',
                'placeholder': 'HH:MM:SS, (example: 00:15:00 for 15 minutes)'
            }),
            'price': forms.NumberInput(attrs={
                'class': 'form-control',
                'placeholder': 'Example: 100.00 (0 for free)'
            }),
            'down_payment': forms.NumberInput(attrs={
                'class': 'form-control',
                'placeholder': 'Example: 50.00 (0 for free)'
            }),
            'image': forms.ClearableFileInput(attrs={'class': 'form-control'}),
            'currency': forms.Select(choices=[('USD', 'USD'), ('EUR', 'EUR'), ('GBP', 'GBP')],
                                     attrs={'class': 'form-control'}),
            'background_color': forms.TextInput(attrs={'class': 'form-control', 'type': 'color', 'value': '#000000'}),
        }



import datetime
from unittest.mock import patch

from django.core.cache import cache
from django.test import override_settings

from appointment.models import StaffMember
from appointment.tests.base.base_test import BaseTest
from appointment.utils.db_helpers import Config, WorkingHours, get_staff_member_buffer_time, \
    get_staff_member_end_time, get_staff_member_slot_duration, get_staff_member_start_time


class BaseStaffMemberTimeTestSetup(BaseTest):
    """Base setup class for staff member time function tests."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()

        # Set staff member-specific settings
        self.staff_member1.slot_duration = 15
        self.staff_member1.lead_time = datetime.time(8, 30)
        self.staff_member1.finish_time = datetime.time(18, 0)
        self.staff_member1.appointment_buffer_time = 45
        self.staff_member1.save()

        # Setting WorkingHours for staff_member1 for Monday
        self.wh = WorkingHours.objects.create(
                staff_member=self.staff_member1,
                day_of_week=1,
                start_time=datetime.time(9, 0),
                end_time=datetime.time(17, 0)
        )

    @override_settings(DEBUG=True)
    def tearDown(self):
        super().tearDown()
        StaffMember.objects.all().delete()
        if Config.objects.exists():
            Config.objects.all().delete()
        WorkingHours.objects.all().delete()
        cache.clear()


@patch('appointment.utils.db_helpers.APPOINTMENT_BUFFER_TIME', 59)
class TestGetStaffMemberBufferTime(BaseStaffMemberTimeTestSetup):
    """Test suite for get_staff_member_buffer_time function."""

    def test_staff_member_buffer_time_with_global_setting(self):
        """Test buffer time when staff member-specific setting is None."""
        self.staff_member1.appointment_buffer_time = None
        self.staff_member1.save()
        buffer_time = get_staff_member_buffer_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(buffer_time, 59)  # Global setting

    def test_staff_member_buffer_time_with_staff_member_setting(self):
        """Test buffer time using staff member-specific setting."""
        buffer_time = get_staff_member_buffer_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(buffer_time, 45)  # Staff member specific setting

    def test_staff_member_buffer_time_with_working_hours_conflict(self):
        """Test buffer time when it conflicts with WorkingHours."""
        self.staff_member1.appointment_buffer_time = 120  # Set a buffer time greater than WorkingHours start time
        self.staff_member1.save()
        buffer_time = get_staff_member_buffer_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(buffer_time, 120)  # Should still use staff member-specific setting even if it causes conflict


@patch('appointment.utils.db_helpers.APPOINTMENT_SLOT_DURATION', 31)
class TestGetStaffMemberSlotDuration(BaseStaffMemberTimeTestSetup):
    """Test suite for get_staff_member_slot_duration function."""

    def test_staff_member_slot_duration_with_global_setting(self):
        """Test slot duration when staff member-specific setting is None."""
        self.staff_member1.slot_duration = None
        self.staff_member1.save()
        slot_duration = get_staff_member_slot_duration(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(slot_duration, 31)  # Global setting

    def test_staff_member_slot_duration_with_staff_member_setting(self):
        """Test slot duration using staff member-specific setting."""
        slot_duration = get_staff_member_slot_duration(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(slot_duration, 15)  # Staff member specific setting


class TestGetStaffMemberStartTime(BaseStaffMemberTimeTestSetup):
    """Test suite for get_staff_member_start_time function."""

    def test_staff_member_start_time(self):
        """Test start time based on WorkingHours."""
        start_time = get_staff_member_start_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(start_time, datetime.time(9, 0))

    def test_staff_member_start_time_with_lead_time(self):
        """Test start time when both lead_time and WorkingHours are available."""
        start_time = get_staff_member_start_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(start_time, self.wh.start_time)  # lead_time should prevail


class TestGetStaffMemberEndTime(BaseStaffMemberTimeTestSetup):
    """Test suite for get_staff_member_end_time function."""

    def test_staff_member_end_time(self):
        """Test end time based on WorkingHours."""
        end_time = get_staff_member_end_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(end_time, datetime.time(17, 0))

    def test_staff_member_end_time_with_finish_time(self):
        """Test end time when both finish_time and WorkingHours are available."""
        end_time = get_staff_member_end_time(self.staff_member1, datetime.date(2023, 10, 9))
        self.assertEqual(end_time, self.wh.end_time)  # finish_time should prevail



import datetime

from appointment.utils.validators import not_in_the_past
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils.translation import gettext as _


class NotInThePastTests(TestCase):
    def test_date_in_the_past_raises_validation_error(self):
        """Test that a date in the past raises a ValidationError."""
        past_date = datetime.date.today() - datetime.timedelta(days=1)
        with self.assertRaises(ValidationError) as context:
            not_in_the_past(past_date)
        self.assertEqual(str(context.exception.message), _('Date is in the past'))

    def test_date_today_does_not_raise_error(self):
        """Test that today's date does not raise an error."""
        today = datetime.date.today()
        try:
            not_in_the_past(today)
        except ValidationError:
            self.fail("not_in_the_past() raised ValidationError unexpectedly for today's date!")

    def test_date_in_the_future_does_not_raise_error(self):
        """Test that a date in the future does not raise an error."""
        future_date = datetime.date.today() + datetime.timedelta(days=1)
        try:
            not_in_the_past(future_date)
        except ValidationError:
            self.fail("not_in_the_past() raised ValidationError unexpectedly for a future date!")



from datetime import timedelta

from django.core.exceptions import ValidationError
from django.utils import timezone

from appointment.models import AppointmentRescheduleHistory
from appointment.tests.base.base_test import BaseTest


class AppointmentRescheduleHistoryCreationTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

    def setUp(self):
        self.appointment_request = self.create_appt_request_for_sm1()
        self.future_date = timezone.now().date() + timedelta(days=3)
        return super().setUp()

    def test_reschedule_history_creation_with_valid_data(self):
        reschedule_history = AppointmentRescheduleHistory.objects.create(
            appointment_request=self.appointment_request,
            date=self.future_date,
            start_time=timezone.now().time(),
            end_time=(timezone.now() + timedelta(hours=1)).time(),
            staff_member=self.staff_member1,
            reason_for_rescheduling="Client request",
            reschedule_status='pending'
        )
        self.assertIsNotNone(reschedule_history)
        self.assertEqual(reschedule_history.reschedule_status, 'pending')
        self.assertTrue(reschedule_history.still_valid())

    def test_auto_generation_of_id_request_on_creation(self):
        reschedule_history = AppointmentRescheduleHistory.objects.create(
            appointment_request=self.appointment_request,
            date=self.future_date,
            start_time=timezone.now().time(),
            end_time=(timezone.now() + timedelta(hours=1)).time(),
            staff_member=self.staff_member1
        )
        self.assertIsNotNone(reschedule_history.id_request)


class AppointmentRescheduleHistoryValidationTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.past_date = timezone.now().date() - timedelta(days=3)
        cls.future_date = timezone.now().date() + timedelta(days=3)

    def setUp(self):
        self.appointment_request = self.create_appt_request_for_sm1()

    def test_creation_with_past_date_raises_validation_error(self):
        with self.assertRaises(ValidationError):
            AppointmentRescheduleHistory.objects.create(
                appointment_request=self.appointment_request,
                date=self.past_date,
                start_time=timezone.now().time(),
                end_time=(timezone.now() + timedelta(hours=1)).time(),
                staff_member=self.staff_member1
            )

    def test_invalid_date_format_raises_type_error(self):
        with self.assertRaises(TypeError):
            AppointmentRescheduleHistory.objects.create(
                appointment_request=self.appointment_request,
                date="invalid-date",
                start_time=timezone.now().time(),
                end_time=(timezone.now() + timedelta(hours=1)).time(),
                staff_member=self.staff_member1
            )


class AppointmentRescheduleHistoryTimingTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        return super().setUpTestData()

    def setUp(self):
        self.appointment_request = self.create_appt_request_for_sm1()
        self.future_date = timezone.now().date() + timedelta(days=3)
        return super().setUp()

    def test_still_valid_within_time_frame(self):
        reschedule_history = AppointmentRescheduleHistory.objects.create(
            appointment_request=self.appointment_request,
            date=self.future_date,
            start_time=timezone.now().time(),
            end_time=(timezone.now() + timedelta(hours=1)).time(),
            staff_member=self.staff_member1,
            reason_for_rescheduling="Client request",
            reschedule_status='pending'
        )
        self.assertTrue(reschedule_history.still_valid())

    def test_still_valid_outside_time_frame(self):
        reschedule_history = AppointmentRescheduleHistory.objects.create(
            appointment_request=self.appointment_request,
            date=self.future_date,
            start_time=timezone.now().time(),
            end_time=(timezone.now() + timedelta(hours=1)).time(),
            staff_member=self.staff_member1,
            reason_for_rescheduling="Client request",
            reschedule_status='pending'
        )
        self.assertTrue(reschedule_history.still_valid())
        # Simulate passage of time beyond the validity window
        reschedule_history.created_at -= timedelta(minutes=6)
        reschedule_history.save()
        self.assertFalse(reschedule_history.still_valid())



# test_view_helpers.py
# Path: appointment/tests/test_view_helpers.py

from django.http import HttpRequest
from django.test import TestCase

from appointment.utils.view_helpers import generate_random_id, get_locale, is_ajax


class GetLocaleTests(TestCase):
    """Test cases for get_locale"""

    def test_get_locale_en(self):
        with self.settings(LANGUAGE_CODE='en'):
            self.assertEqual(get_locale(), 'en')

    def test_get_locale_en_us(self):
        with self.settings(LANGUAGE_CODE='en_US'):
            self.assertEqual(get_locale(), 'en')

    def test_get_locale_fr(self):
        # Set the local to French
        with self.settings(LANGUAGE_CODE='fr'):
            self.assertEqual(get_locale(), 'fr')

    def test_get_locale_fr_France(self):
        # Set the local to French
        with self.settings(LANGUAGE_CODE='fr_FR'):
            self.assertEqual(get_locale(), 'fr')

    def test_get_locale_others(self):
        with self.settings(LANGUAGE_CODE='de'):
            self.assertEqual(get_locale(), 'de')


class IsAjaxTests(TestCase):
    """Test cases for is_ajax"""

    def test_is_ajax_true(self):
        request = HttpRequest()
        request.META['HTTP_X_REQUESTED_WITH'] = 'XMLHttpRequest'
        self.assertTrue(is_ajax(request))

    def test_is_ajax_false(self):
        request = HttpRequest()
        self.assertFalse(is_ajax(request))


class GenerateRandomIdTests(TestCase):
    """Test cases for generate_random_id"""

    def test_generate_random_id(self):
        id1 = generate_random_id()
        id2 = generate_random_id()
        self.assertNotEqual(id1, id2)



# error_codes.py
# Path: appointment/utils/error_codes.py


"""
Author: Adams Pierre David
Since: 2.0.0
"""

from enum import Enum, auto


class ErrorCode(Enum):
    APPOINTMENT_CONFLICT = auto()
    APPOINTMENT_NOT_FOUND = auto()
    DAY_OFF_CONFLICT = auto()
    DAY_OFF_NOT_FOUND = auto()
    INVALID_DATA = auto()
    INVALID_DATE = auto()
    NOT_AUTHORIZED = auto()
    PAST_DATE = auto()
    STAFF_ID_REQUIRED = auto()
    WORKING_HOURS_NOT_FOUND = auto()
    WORKING_HOURS_CONFLICT = auto()
    SERVICE_NOT_FOUND = auto()
    STAFF_MEMBER_NOT_FOUND = auto()



# db_helpers.py
# Path: appointment/utils/db_helpers.py

"""
Author: Adams Pierre David
Since: 2.0.0
"""

import datetime
from typing import Optional
from urllib.parse import urlparse

from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.core.exceptions import FieldDoesNotExist
from django.urls import reverse
from django.utils import timezone

from appointment.logger_config import get_logger
from appointment.settings import (
    APPOINTMENT_BUFFER_TIME, APPOINTMENT_FINISH_TIME, APPOINTMENT_LEAD_TIME, APPOINTMENT_PAYMENT_URL,
    APPOINTMENT_SLOT_DURATION, APPOINTMENT_WEBSITE_NAME
)
from appointment.utils.date_time import combine_date_and_time, get_weekday_num

logger = get_logger(__name__)

# Check if django-q is installed in settings
DJANGO_Q_AVAILABLE = 'django_q' in settings.INSTALLED_APPS

# Check if django-q is installed as a dependency
try:
    from django_q.models import Schedule
    from django_q.tasks import schedule

    DJANGO_Q_AVAILABLE = True
except ImportError:
    DJANGO_Q_AVAILABLE = False
    Schedule = None
    schedule = None
    logger.warning("django-q is not installed. Email reminders will not be scheduled.")

Appointment = apps.get_model('appointment', 'Appointment')
AppointmentRequest = apps.get_model('appointment', 'AppointmentRequest')
WorkingHours = apps.get_model('appointment', 'WorkingHours')
DayOff = apps.get_model('appointment', 'DayOff')
PaymentInfo = apps.get_model('appointment', 'PaymentInfo')
StaffMember = apps.get_model('appointment', 'StaffMember')
Config = apps.get_model('appointment', 'Config')
Service = apps.get_model('appointment', 'Service')
EmailVerificationCode = apps.get_model('appointment', 'EmailVerificationCode')
AppointmentRescheduleHistory = apps.get_model('appointment', 'AppointmentRescheduleHistory')


def calculate_slots(start_time, end_time, buffer_time, slot_duration):
    """Calculate the available slots between the given start and end times using the given buffer time and slot duration

    :param start_time: The start time.
    :param end_time: The end time.
    :param buffer_time: The buffer time.
    :param slot_duration: The duration of each slot.
    :return: A list of available slots.
    """
    slots = []
    buffer_time = buffer_time.replace(tzinfo=None)
    while start_time + slot_duration <= end_time:
        if start_time >= buffer_time:
            slots.append(start_time)
        start_time += slot_duration
    return slots


def calculate_staff_slots(date, staff_member):
    """Calculate the available slots for the given staff member on the given date.

    :param date: The date to calculate the slots for.
    :param staff_member: The staff member to calculate the slots for.
    :return: A list of available slots.
    """
    # Convert the times to datetime objects
    weekday_num = get_weekday_num_from_date(date)
    if not is_working_day(staff_member, weekday_num):
        return []
    staff_member_start_time = get_staff_member_start_time(staff_member, date)
    start_time = datetime.datetime.combine(date, staff_member_start_time)
    end_time = datetime.datetime.combine(date, get_staff_member_end_time(staff_member, date))

    # Convert the buffer duration in minutes to a timedelta object
    buffer_duration_minutes = get_staff_member_buffer_time(staff_member, date)
    buffer_duration = datetime.timedelta(minutes=buffer_duration_minutes)
    buffer_time_init = datetime.datetime.combine(date, staff_member_start_time)
    buffer_time = buffer_time_init + buffer_duration

    # Convert slot duration to a timedelta object
    slot_duration_minutes = get_staff_member_slot_duration(staff_member, date)
    slot_duration = datetime.timedelta(minutes=slot_duration_minutes)

    return calculate_slots(start_time, end_time, buffer_time, slot_duration)


def check_day_off_for_staff(staff_member, date) -> bool:
    """Check if the given staff member is off on the given date.
    :param staff_member: The staff member to check.
    :param date: The date to check.
    """
    return DayOff.objects.filter(staff_member=staff_member, start_date__lte=date, end_date__gte=date).exists()


def create_and_save_appointment(ar, client_data: dict, appointment_data: dict, request):
    """Create and save a new appointment based on the provided appointment request and client data.

    :param ar: The appointment request associated with the new appointment.
    :param client_data: The data of the client making the appointment.
    :param appointment_data: Additional data for the appointment, including phone number, address, etc.
    :param request: The request object.
    :return: The newly created appointment.
    """
    user = get_user_by_email(client_data['email'])
    appointment = Appointment.objects.create(
            client=user, appointment_request=ar,
            **appointment_data
    )
    appointment.save()
    logger.info(f"New appointment created: {appointment.to_dict()}")
    if appointment.want_reminder:
        logger.info(f"User wants a reminder for appointment {appointment.id}, scheduling it...")
        if DJANGO_Q_AVAILABLE:
            schedule_email_reminder(appointment, request)
        else:
            logger.warning(f"Email reminder requested for appointment {appointment.id}, but django-q is not available.")
    return appointment


def schedule_email_reminder(appointment, request, appointment_datetime=None):
    """Schedule an email reminder for the given appointment."""
    if not DJANGO_Q_AVAILABLE:
        logger.warning("Django-Q is not available. Email reminder will not be scheduled.")
        return

    # Calculate reminder datetime if not provided
    if appointment_datetime is None:
        appointment_datetime = combine_date_and_time(appointment.appointment_request.date,
                                                     appointment.appointment_request.start_time)
    if timezone.is_naive(appointment_datetime):
        appointment_datetime = timezone.make_aware(appointment_datetime)

    reminder_datetime = appointment_datetime - datetime.timedelta(days=1)

    ar_id_request = appointment.appointment_request.get_id_request()
    relative_reschedule_url = reverse('appointment:prepare_reschedule_appointment', args=[ar_id_request])
    reschedule_link = get_absolute_url_(relative_reschedule_url, request)

    logger.info(f"Scheduling email reminder for appointment {appointment.id} at {reminder_datetime}")

    # Schedule the email reminder task with Django-Q
    schedule('appointment.tasks.send_email_reminder',
             to_email=appointment.client.email,
             name=f"reminder_{appointment.id_request}",
             first_name=appointment.client.first_name,
             reschedule_link=reschedule_link,
             appointment_id=appointment.id,
             schedule_type=Schedule.ONCE,  # Use Schedule.ONCE for a one-time task
             next_run=reminder_datetime)


def update_appointment_reminder(appointment, new_date, new_start_time, request, want_reminder=None):
    """
    Updates or cancels the appointment reminder based on changes to the start time or date,
    and the user's preference for receiving a reminder.
    """
    if not DJANGO_Q_AVAILABLE:
        logger.warning("Django-Q is not available. Appointment reminder cannot be updated.")
        return
    # Convert new date and time strings to datetime objects for comparison
    new_datetime = combine_date_and_time(new_date, new_start_time)

    existing_datetime = combine_date_and_time(appointment.appointment_request.date,
                                              appointment.appointment_request.start_time)

    # Ensure new_datetime is timezone-aware
    if timezone.is_naive(new_datetime):
        new_datetime = timezone.make_aware(new_datetime)

    # Ensure existing_datetime is timezone-aware
    if timezone.is_naive(existing_datetime):
        existing_datetime = timezone.make_aware(existing_datetime)

    # Determine if there's been a change in the datetime or the reminder preference
    want_reminder = want_reminder if want_reminder is not None else appointment.want_reminder
    datetime_changed = new_datetime != existing_datetime
    reminder_preference_changed = appointment.want_reminder != want_reminder

    if datetime_changed or reminder_preference_changed:
        # Cancel any existing reminder
        cancel_existing_reminder(appointment.id_request)

        # If a reminder is still desired and the appointment is in the future, schedule a new one
        if want_reminder and new_datetime > timezone.now():
            schedule_email_reminder(appointment, request, new_datetime)
        else:
            logger.info(
                    f"Reminder for appointment {appointment.id} is not scheduled per "
                    f"user's preference or past datetime.")

    # Update the appointment's reminder preference
    appointment.want_reminder = want_reminder
    appointment.save()


def cancel_existing_reminder(appointment_id_request):
    """
    Cancels any existing reminder for the appointment.
    """
    if not DJANGO_Q_AVAILABLE:
        logger.warning("Django-Q is not available. Appointment reminder cannot be updated.")
        return
    task_name = f"reminder_{appointment_id_request}"
    Schedule.objects.filter(name=task_name).delete()


def can_appointment_be_rescheduled(appointment_request):
    # Datetime 5 minutes ago from now
    five_minutes_ago = timezone.now() - timezone.timedelta(minutes=5)

    # Filter reschedule histories to those created within the last 5 minutes
    recent_reschedule_count = appointment_request.reschedule_histories.filter(created_at__gte=five_minutes_ago).count()
    service = appointment_request.service
    config = Config.get_instance()

    # Determine which rescheduled limit to use based on service settings
    if service.allow_rescheduling:
        # If rescheduling is allowed
        logger.info(f"Rescheduling is allowed for service {service.name} -> "
                    f"Reschedule count: {recent_reschedule_count}, Reschedule limit: {service.reschedule_limit}")
        return recent_reschedule_count < service.reschedule_limit
    else:
        # Rescheduling is allowed but no specific limit set; use system default
        logger.info(f"Rescheduling is allowed but no specific limit set for service {service.name} -> "
                    f"Reschedule count: {recent_reschedule_count}, Reschedule limit: {config.default_reschedule_limit}")
        return recent_reschedule_count < config.default_reschedule_limit


def staff_change_allowed_on_reschedule():
    return Config.objects.first().allow_staff_change_on_reschedule


def generate_unique_username_from_email(email: str) -> str:
    username_base = email.split('@')[0]
    username = username_base
    suffix = 1
    CLIENT_MODEL = get_user_model()

    while CLIENT_MODEL.objects.filter(username=username).exists():
        suffix_str = f"{suffix:02}"
        username = f"{username_base}{suffix_str}"
        suffix += 1
    return username


def parse_name(name: str):
    parts = name.split(' ', 1)
    if len(parts) == 1:
        parts.append('')  # Add an empty string for the last name if not provided
    return parts[0], parts[1]


def create_user_with_email(client_data: dict):
    CLIENT_MODEL = get_user_model()
    # Valid fields
    valid_fields = ['email', 'first_name', 'last_name']

    # Filter client_data to include only valid fields
    user_data = {field: client_data.get(field, '') for field in valid_fields}

    user = CLIENT_MODEL.objects.create_user(**user_data)
    return user


def create_user_with_username(client_data: dict):
    CLIENT_MODEL = get_user_model()
    if 'username' not in client_data:
        username = generate_unique_username_from_email(client_data['email'])
    else:
        username = client_data['username']
    user_data = {
        'username': username,
        'email': client_data['email'],
        'first_name': client_data.get('first_name', ''),
        'last_name': client_data.get('last_name', '')
    }
    user = CLIENT_MODEL.objects.create_user(**user_data)
    return user


def create_new_user(client_data: dict):
    CLIENT_MODEL = get_user_model()

    # Check if client_data has first_name and last_name
    if 'first_name' not in client_data or 'last_name' not in client_data:
        # Assuming 'name' contains a single space between first and last name.
        client_data['first_name'], client_data['last_name'] = parse_name(client_data['name'])

    try:
        # Check if the 'username' field exists in the User model
        CLIENT_MODEL._meta.get_field('username')
        user = create_user_with_username(client_data)
    except FieldDoesNotExist:
        client_data.pop('username', None)
        user = create_user_with_email(client_data)
    return user


def username_in_user_model():
    CLIENT_MODEL = get_user_model()
    try:
        # Check if the 'username' field exists in the User model
        CLIENT_MODEL._meta.get_field('username')
        return True
    except FieldDoesNotExist:
        return False


def create_payment_info_and_get_url(appointment):
    """
    Create a new payment information entry for the appointment and return the payment URL.

    :param appointment: The appointment to create the payment information for.
    :return: The payment URL for the appointment.
    """
    # Create a new PaymentInfo entry for the appointment
    payment_info = PaymentInfo(appointment=appointment)
    payment_info.save()

    # Check if APPOINTMENT_PAYMENT_URL is a Django reverse URL (e.g. "app:view_name") or an external link
    if (":" in APPOINTMENT_PAYMENT_URL and "/" not in APPOINTMENT_PAYMENT_URL) and not bool(
            urlparse(APPOINTMENT_PAYMENT_URL).netloc):
        # It's a Django reverse URL; generate the URL
        payment_url = reverse(
                APPOINTMENT_PAYMENT_URL,
                kwargs={'object_id': payment_info.id, 'id_request': payment_info.get_id_request()}
        )
    else:
        # It's an external link; return as is or append necessary data
        payment_url = APPOINTMENT_PAYMENT_URL

    return payment_url


def exclude_booked_slots(appointments, slots, slot_duration=None):
    """Exclude the booked slots from the given list of slots.

    :param appointments: The appointments to exclude.
    :param slots: The slots to exclude the appointments from.
    :param slot_duration: The duration of each slot.
    :return: The slots with the booked slots excluded.
    """
    available_slots = []
    for slot in slots:
        slot_end = slot + slot_duration
        is_available = True
        for appointment in appointments:
            appointment_start_time = appointment.get_start_time()
            appointment_end_time = appointment.get_end_time()
            if appointment_start_time < slot_end and slot < appointment_end_time:
                is_available = False
                break
        if is_available:
            available_slots.append(slot)
    return available_slots


def exclude_pending_reschedules(slots, staff_member, date):
    """
    Exclude the slots that are pending reschedule for the given staff member and date.
    """

    # Calculate the time window for "last 5 minutes"
    ten_minutes_ago = timezone.now() - datetime.timedelta(minutes=5)
    pending_reschedules = AppointmentRescheduleHistory.objects.filter(
            appointment_request__staff_member=staff_member,
            date=date,
            reschedule_status='pending',
            created_at__gte=ten_minutes_ago
    )

    # Filter out slots that overlap with any pending rescheduling
    filtered_slots = slots[:]
    for reschedule in pending_reschedules:
        reschedule_start_time = datetime.datetime.combine(date, reschedule.start_time)
        reschedule_end_time = datetime.datetime.combine(date, reschedule.end_time)

        filtered_slots = [slot for slot in filtered_slots if
                          not (reschedule_start_time <= slot < reschedule_end_time)]

    return filtered_slots


def day_off_exists_for_date_range(staff_member, start_date, end_date, days_off_id=None) -> bool:
    """Check if a day off exists for the given staff member and date range.

    :param staff_member: The staff member to check.
    :param start_date: The start date of the date range.
    :param end_date: The end date of the date range.
    :param days_off_id: The ID of the day off to exclude from the check.
    :return: True if a day off exists for the given staff member and date range; otherwise, False.
    """
    days_off = DayOff.objects.filter(staff_member=staff_member, start_date__lte=end_date, end_date__gte=start_date)
    if days_off_id:
        days_off = days_off.exclude(id=days_off_id)
    return days_off.exists()


def get_all_appointments() -> list:
    """Get all appointments from the database.

    :return: QuerySet, all appointments
    """
    return Appointment.objects.all()


def get_all_staff_members() -> list:
    """Get all staff members from the database.

    :return: QuerySet, all staff members
    """
    return StaffMember.objects.all()


def get_appointment_buffer_time():
    """Get the appointment buffer time from the settings file.

    :return: The appointment buffer time
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.appointment_buffer_time:
        return config.appointment_buffer_time
    return APPOINTMENT_BUFFER_TIME


def get_appointment_by_id(appointment_id):
    """Get an appointment by its ID.

    :param appointment_id: The appointment's ID
    :return: The appointment with the specified ID or None if no appointment with the specified ID exists
    """
    try:
        return Appointment.objects.get(id=appointment_id)
    except Appointment.DoesNotExist:
        return None


def get_appointment_finish_time():
    """Get the appointment finish time from the settings file.

    :return: The appointment's finish time
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.finish_time:
        return config.finish_time
    return APPOINTMENT_FINISH_TIME


def get_appointment_lead_time():
    """Get the appointment lead time from the settings file.

    :return: The appointment's lead time
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.lead_time:
        return config.lead_time
    return APPOINTMENT_LEAD_TIME


def get_appointment_slot_duration():
    """Get the appointment slot duration from the settings file.

    :return: The appointment slot duration
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.slot_duration:
        return config.slot_duration
    return APPOINTMENT_SLOT_DURATION


def get_appointments_for_date_and_time(date, start_time, end_time, staff_member):
    """Returns all appointments that overlap with the specified date and time range.

    :param date: The date to filter appointments on.
    :param start_time: The starting time to filter appointments on.
    :param end_time: The ending time to filter appointments on.
    :param staff_member: The staff member to filter appointments on.

    :return: QuerySet, all appointments that overlap with the specified date and time range
    """
    return Appointment.objects.filter(
            appointment_request__date=date,
            appointment_request__start_time__lte=end_time,
            appointment_request__end_time__gte=start_time,
            appointment_request__staff_member=staff_member
    )


def get_config():
    """Returns the configuration object from the database or the cache."""
    config = cache.get('config')
    if not config:
        config = Config.objects.first()
        # Cache the configuration for 1 hour (3600 seconds)
        cache.set('config', config, 3600)
    return config


def get_day_off_by_id(day_off_id):
    """Get a day off by its ID.

    :param day_off_id: The day offs ID
    :return: DayOff, the day off with the specified ID or None if no day off with the specified ID exists.
    """
    try:
        return DayOff.objects.get(pk=day_off_id)
    except DayOff.DoesNotExist:
        return None


def get_non_working_days_for_staff(staff_member_id):
    """Return the non-working days for the given staff member or an empty list if the staff member does not exist."""
    all_days = set(range(7))  # Represents all days (0-6)
    try:
        staff_member = StaffMember.objects.get(id=staff_member_id)
        working_days = set(WorkingHours.objects.filter(staff_member=staff_member).values_list('day_of_week', flat=True))

        # Subtracting working_days from all_days to get non-working days
        non_working_days = list(all_days - working_days)
        return non_working_days
    except StaffMember.DoesNotExist:
        return []


def get_staff_member_appointment_list(staff_member: StaffMember) -> list:
    """Get a list of appointments for the given staff member."""
    return Appointment.objects.filter(appointment_request__staff_member=staff_member)


def get_weekday_num_from_date(date: datetime.date = None) -> int:
    """Get the number of the weekday from the given date."""
    if date is None:
        date = datetime.date.today()
    return get_weekday_num(date.strftime("%A"))


def get_staff_member_buffer_time(staff_member: StaffMember, date: datetime.date) -> float:
    """Return the buffer time for the given staff member on the given date."""
    _, _, _, buff_time = get_times_from_config(date)
    buffer_minutes = buff_time.total_seconds() / 60
    return staff_member.appointment_buffer_time or buffer_minutes


def get_staff_member_by_user_id(user_id):
    """Return a staff member by their user ID."""
    try:
        return StaffMember.objects.get(user_id=user_id)
    except StaffMember.DoesNotExist:
        return None


def get_staff_member_end_time(staff_member: StaffMember, date: datetime.date) -> Optional[datetime.time]:
    """Return the end time for the given staff member on the given date."""
    weekday_num = get_weekday_num_from_date(date)
    working_hours = get_working_hours_for_staff_and_day(staff_member, weekday_num)
    return working_hours['end_time']


def get_staff_member_from_user_id_or_logged_in(user, user_id=None):
    """Fetch StaffMember based on the user_id or the logged-in user."""
    staff_member = None
    try:
        if user_id:
            staff_member = StaffMember.objects.get(user_id=user_id)
        else:
            staff_member = user.staffmember
    except StaffMember.DoesNotExist:
        pass
    return staff_member


def get_staff_member_slot_duration(staff_member: StaffMember, date: datetime.date) -> int:
    """Return the slot duration for the given staff member on the given date."""
    _, _, slot_duration, _ = get_times_from_config(date)
    slot_minutes = slot_duration.total_seconds() / 60
    return staff_member.slot_duration or slot_minutes


def get_staff_member_start_time(staff_member: StaffMember, date: datetime.date) -> Optional[datetime.time]:
    """Return the start time for the given staff member on the given date."""
    weekday_num = get_weekday_num_from_date(date)
    working_hours = get_working_hours_for_staff_and_day(staff_member, weekday_num)
    return working_hours['start_time']


def get_times_from_config(date):
    """Get the start time, end time, slot duration, and buffer time from the configuration or the settings file.

    :param date: The date to get the times for.
    :return: The start time, end time, slot duration, and buffer time.
    """
    config = get_config()
    if config:
        start_time = datetime.datetime.combine(date, datetime.time(hour=config.lead_time.hour,
                                                                   minute=config.lead_time.minute))
        end_time = datetime.datetime.combine(date, datetime.time(hour=config.finish_time.hour,
                                                                 minute=config.finish_time.minute))
        slot_duration = datetime.timedelta(minutes=config.slot_duration)
        buff_time = datetime.timedelta(minutes=config.appointment_buffer_time)
    else:
        start_hour, start_minute = APPOINTMENT_LEAD_TIME
        start_time = datetime.datetime.combine(date, datetime.time(hour=start_hour, minute=start_minute))
        finish_hour, finish_minute = APPOINTMENT_FINISH_TIME
        end_time = datetime.datetime.combine(date, datetime.time(hour=finish_hour, minute=finish_minute))
        slot_duration = datetime.timedelta(minutes=APPOINTMENT_SLOT_DURATION)
        buff_time = datetime.timedelta(minutes=APPOINTMENT_BUFFER_TIME)
    return start_time, end_time, slot_duration, buff_time


def get_user_by_email(email: str):
    """Get a user by their email address.

    :param email: The email address of the user.
    :return: The user with the specified email address, if found; otherwise, None.
    """
    CLIENT_MODEL = get_user_model()
    return CLIENT_MODEL.objects.filter(email=email).first()


def get_website_name() -> str:
    """Get the website name from the configuration file.

    :return: The website name
    """
    from appointment.models import Config

    config = Config.objects.first()

    if config and config.website_name != "":
        return config.website_name
    return APPOINTMENT_WEBSITE_NAME


def get_working_hours_by_id(working_hours_id):
    """Get a working hours by its ID.

    :param working_hours_id: The working hours' ID
    :return: WorkingHours, the working hours with the specified ID
    """
    try:
        return WorkingHours.objects.get(pk=working_hours_id)
    except WorkingHours.DoesNotExist:
        return None


def get_working_hours_for_staff_and_day(staff_member, day_of_week):
    """Get the working hours for the given staff member and day of the week.

    :param staff_member: The staff member to get the working hours for.
    :param day_of_week: The day of the week to get the working hours for.
    :return: The working hours for the given staff member and day of the week.
    """
    working_hours = WorkingHours.objects.filter(staff_member=staff_member, day_of_week=day_of_week).first()
    start_time = staff_member.get_lead_time()
    end_time = staff_member.get_finish_time()
    if not working_hours and not (start_time and end_time):
        return None
    # If no specific working hours are set for that day, use the default start and end times from StaffMember
    if not working_hours:
        return {
            'staff_member': staff_member,
            'day_of_week': day_of_week,
            'start_time': staff_member.get_lead_time(),
            'end_time': staff_member.get_finish_time()
        }

    # If a WorkingHours instance is found, convert it to a dictionary for consistent return type
    return {
        'staff_member': working_hours.staff_member,
        'day_of_week': working_hours.day_of_week,
        'start_time': working_hours.start_time,
        'end_time': working_hours.end_time
    }


def is_working_day(staff_member: StaffMember, day: int) -> bool:
    """Check if the given day is a working day for the staff member."""
    working_days = list(WorkingHours.objects.filter(staff_member=staff_member).values_list('day_of_week', flat=True))
    return day in working_days


def working_hours_exist(day_of_week, staff_member):
    """Check if working hours exist for the given day of the week and staff member."""
    return WorkingHours.objects.filter(day_of_week=day_of_week, staff_member=staff_member).exists()


def get_absolute_url_(relative_url, request):
    return request.build_absolute_uri(relative_url)



# apps.py
# Path: appointment/apps.py

"""
Author: Adams Pierre David
Since: 1.0.0
"""

from django.apps import AppConfig


class AppointmentConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "appointment"



# test_json_context.py
# Path: appointment/tests/utils/test_json_context.py

import json

from django.test import RequestFactory

from appointment.tests.base.base_test import BaseTest
from appointment.utils.json_context import (
    convert_appointment_to_json, get_generic_context, get_generic_context_with_extra, handle_unauthorized_response,
    json_response
)


class ConvertAppointmentToJsonTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.appointments = [self.create_appt_for_sm1()]
        self.request = self.factory.get('/')
        self.request.user = self.users['client1']

    def test_convert_appointment_to_json(self):
        """Test if an appointment can be converted to JSON."""
        data = convert_appointment_to_json(self.request, self.appointments)
        self.assertIsInstance(data, list, "Data should be a list")
        self.assertEqual(len(data), 1, "Data list should have one appointment")
        self.assertIn("id", data[0], "Data should contain 'id' field")


class JsonResponseTests(BaseTest):
    def test_json_response(self):
        """Test if a JSON response can be created."""
        message = "Gate Room Under Attack"
        response = json_response(message=message)
        self.assertEqual(response.status_code, 200, "Response status should be 200")
        content = json.loads(response.content.decode('utf-8'))
        self.assertEqual(content['message'], message, "Response content should match the message")


class GetGenericContextTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.user1 = self.users['client1']
        self.request = self.factory.get('/')
        self.request.user = self.user1

    def test_get_generic_context(self):
        """Test if a generic context can be created."""
        context = get_generic_context(self.request)
        self.assertEqual(context['user'], self.user1, "Context user should match the request user")
        self.assertIn('BASE_TEMPLATE', context, "Context should contain 'BASE_TEMPLATE'")
        self.assertIn('is_superuser', context, "Context should contain 'is_superuser'")


class GetGenericContextWithExtraTests(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.user1 = self.users['client1']
        self.request = self.factory.get('/')
        self.request.user = self.user1
        self.extra = {"key": "value"}

    def test_get_generic_context_with_extra(self):
        """Test if a generic context with extra data can be created."""
        context = get_generic_context_with_extra(self.request, self.extra)
        self.assertEqual(context['user'], self.user1, "Context user should match the request user")
        self.assertEqual(context['key'], "value", "Context should include extra data")
        self.assertIn('BASE_TEMPLATE', context, "Context should contain 'BASE_TEMPLATE'")
        self.assertIn('is_superuser', context, "Context should contain 'is_superuser'")


class HandleUnauthorizedResponseTests(BaseTest):
    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()
        self.message = "Unauthorized"

    def test_handle_unauthorized_response_json(self):
        """Test if an unauthorized response can be created when the response type is JSON."""
        request = self.factory.get('/')
        response = handle_unauthorized_response(request=request, message=self.message, response_type='json')
        self.assertEqual(response.status_code, 403, "Response status should be 403")
        content = json.loads(response.content.decode('utf-8'))
        self.assertEqual(content['message'], self.message, "Response content should match the message")

    def test_handle_unauthorized_response_html(self):
        """Test if an unauthorized response can be created when the response type is HTML."""
        request = self.factory.get('/app-admin/user-events/')
        response = handle_unauthorized_response(request, self.message, 'html')
        self.assertEqual(response.status_code, 403, "Response status should be 403")



import datetime
from copy import deepcopy
from datetime import timedelta

from django.db import IntegrityError
from django.utils.translation import gettext as _

from appointment.models import Config, DayOff, Service, StaffMember, WorkingHours
from appointment.tests.base.base_test import BaseTest


class StaffMemberCreationTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.staff_member = cls.staff_member1
        cls.staff = cls.users['staff1']
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_default_attributes_on_creation(self):
        self.assertIsNotNone(self.staff_member)
        self.assertEqual(self.staff_member.user, self.staff)
        self.assertEqual(list(self.staff_member.get_services_offered()), [self.service])
        self.assertIsNone(self.staff_member.lead_time)
        self.assertIsNone(self.staff_member.finish_time)
        self.assertIsNone(self.staff_member.slot_duration)
        self.assertIsNone(self.staff_member.appointment_buffer_time)
        self.assertIsNotNone(self.staff_member.created_at)
        expected_str = self.staff_member.get_staff_member_name()
        self.assertEqual(str(self.staff_member), expected_str)

    def test_creation_without_service(self):
        """A staff member can be created without a service."""
        new_staff = self.create_user_(first_name="Jonas", last_name="Quinn", email="jonas.quinn@django-appointment.com",
                                      username="jonas.quinn")
        new_staff_member = StaffMember.objects.create(user=new_staff)
        self.assertIsNotNone(new_staff_member)
        self.assertEqual(new_staff_member.services_offered.count(), 0)

    def test_creation_fails_without_user(self):
        """A staff member cannot be created without a user."""
        with self.assertRaises(IntegrityError):
            StaffMember.objects.create()

    def test_get_staff_member_name_with_email(self):
        # Simulate create a staff member with only an email and username
        # (in my case, username is mandatory, but should work with email as well)
        email_only_user = self.create_user_(
                first_name="",
                last_name="",
                email="mckay.rodney@django-appointment.com",
                username="mckay.rodney@django-appointment.com"
        )
        email_only_staff = StaffMember.objects.create(user=email_only_user)

        # Test that the email is returned when no other name information is available
        self.assertEqual(email_only_staff.get_staff_member_name(), "mckay.rodney@django-appointment.com")

    def test_get_staff_member_name_priority(self):
        # Create a staff member with all name fields filled
        full_user = self.create_user_(
                first_name="Rodney",
                last_name="McKay",
                email="mckay.rodney@django-appointment.com",
                username="rodney.mckay"
        )
        full_staff = StaffMember.objects.create(user=full_user)

        # Test that the full name is returned when available
        self.assertEqual(full_staff.get_staff_member_name(), "Rodney McKay")

        # Modify the user to test different scenarios
        full_user.first_name = ""
        full_user.last_name = ""
        full_user.save()

        # Test that the username is returned when full name is not available
        self.assertEqual(full_staff.get_staff_member_name(), "rodney.mckay")

        full_user.username = ""
        full_user.save()

        # Test that the email is returned when username and full name are not available
        self.assertEqual(full_staff.get_staff_member_name(), "mckay.rodney@django-appointment.com")


class StaffMemberServiceTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.staff_member = cls.staff_member1
        cls.staff = cls.users['staff1']
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def get_fresh_staff_member(self):
        return deepcopy(self.staff_member)

    def test_association_with_multiple_services(self):
        """A staff member can offer multiple services."""
        ori = self.create_service_(name="Ori Shield Configuration", duration=timedelta(hours=1), price=100000,
                                   description="Configure the Ori shield")
        symbiote = self.create_service_(name="Symbiote Extraction", duration=timedelta(hours=1), price=100000,
                                        description="Extract a symbiote")
        sm = self.get_fresh_staff_member()

        sm.services_offered.add(ori)
        sm.services_offered.add(symbiote)

        self.assertIn(ori, sm.services_offered.all())
        self.assertIn(symbiote, sm.services_offered.all())

    def test_services_offered(self):
        """Test get_services_offered & get_service_offered_text function."""
        self.assertIn(self.service, self.staff_member.get_services_offered())
        self.assertEqual(self.staff_member.get_service_offered_text(), self.service.name)
        self.assertTrue(self.staff_member.get_service_is_offered(self.service.pk))

    def test_staff_member_with_non_existent_service(self):
        """A staff member cannot offer a non-existent service."""
        new_staff = self.create_user_(first_name="Vala", last_name="Mal Doran",
                                      email="vala.mal-doran@django-appointment.com", username="vala.mal-doran")
        new_staff_member = StaffMember.objects.create(user=new_staff)

        # Trying to add a non-existent service to the staff member's services_offered
        with self.assertRaises(ValueError):
            new_staff_member.services_offered.add(
                    Service(id=9999, name="Prometheus Acquisition", duration=timedelta(hours=2), price=200))


class StaffMemberWorkingTimeTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.staff_member = cls.staff_member1
        cls.staff = cls.users['staff1']
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def get_fresh_staff_member(self):
        return deepcopy(self.staff_member)

    def test_working_both_weekend_days(self):
        """Test works_on_both_weekends_day method."""
        sm = self.get_fresh_staff_member()
        sm.work_on_saturday = True
        sm.work_on_sunday = True
        self.assertTrue(sm.works_on_both_weekends_day())

    def test_get_non_working_days(self):
        """Test get_non_working_days method."""
        sm = self.get_fresh_staff_member()
        sm.work_on_saturday = False
        sm.work_on_sunday = False
        self.assertEqual(sm.get_non_working_days(), [6, 0])  # [6, 0] represents Saturday and Sunday respectively

    def test_identification_of_all_non_working_days(self):
        """Test various combinations of weekend work using a deepcopy of staff_member."""
        # Test with work on Saturday only
        sm = self.get_fresh_staff_member()

        sm.work_on_saturday = True
        sm.work_on_sunday = False
        sm.save()
        self.assertEqual(sm.get_weekend_days_worked_text(), _("Saturday"))

        # Test with work on both Saturday and Sunday
        sm.work_on_saturday = True
        sm.work_on_sunday = True
        sm.save()
        self.assertEqual(sm.get_weekend_days_worked_text(), _("Saturday and Sunday"))

        # Test with work on Sunday only
        sm.work_on_saturday = False
        sm.work_on_sunday = True
        sm.save()
        self.assertEqual(sm.get_weekend_days_worked_text(), _("Sunday"))

        # Test with work on neither Saturday nor Sunday
        sm.work_on_saturday = False
        sm.work_on_sunday = False
        sm.save()
        self.assertEqual(sm.get_weekend_days_worked_text(), _("None"))

    def test_get_days_off(self):
        """Test retrieval of days off."""
        sm = self.get_fresh_staff_member()
        DayOff.objects.create(staff_member=sm, start_date="2023-01-01", end_date="2023-01-02")
        self.assertEqual(len(sm.get_days_off()), 1)

    def test_get_working_hours(self):
        """Test retrieval of working hours."""
        sm = self.get_fresh_staff_member()
        WorkingHours.objects.create(staff_member=sm, day_of_week=1, start_time=datetime.time(9, 0),
                                    end_time=datetime.time(17, 0))
        self.assertEqual(len(sm.get_working_hours()), 1)

        # Precautionary cleanup (FIRST principle)
        WorkingHours.objects.all().delete()

    def test_update_upon_working_hours_deletion(self):
        """Test the update of work_on_saturday and work_on_sunday upon working-hours deletion."""
        sm = self.get_fresh_staff_member()
        sm.work_on_saturday = True
        sm.work_on_sunday = True
        sm.save()

        sm.update_upon_working_hours_deletion(6)
        self.assertFalse(sm.work_on_saturday)
        sm.update_upon_working_hours_deletion(0)
        self.assertFalse(sm.work_on_sunday)

    def test_is_working_day(self):
        """Test whether a day is considered a working day."""
        sm = self.get_fresh_staff_member()
        sm.work_on_saturday = False
        sm.work_on_sunday = False
        sm.save()

        self.assertFalse(self.staff_member.is_working_day(6))  # Saturday
        self.assertTrue(self.staff_member.is_working_day(1))  # Monday


class StaffMemberGetterTestCase(BaseTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.staff_member = cls.staff_member1
        cls.staff = cls.users['staff1']
        cls.service = cls.service1

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def get_fresh_staff_member(self):
        return deepcopy(self.staff_member)

    def test_get_staff_member_first_name(self):
        """Test that the staff member's first name is returned."""
        self.assertEqual(self.staff_member.get_staff_member_first_name(), self.staff.first_name)

    def test_config_values_takes_over_when_sm_values_null(self):
        """When some values are null in the StaffMember, the Config values should be used."""
        config = Config.objects.create(
                lead_time=datetime.time(9, 34),
                finish_time=datetime.time(17, 11),
                slot_duration=37,
                appointment_buffer_time=16
        )
        # Checking that the StaffMember's values are None
        self.assertIsNone(self.staff_member.slot_duration)
        self.assertIsNone(self.staff_member.lead_time)
        self.assertIsNone(self.staff_member.finish_time)
        self.assertIsNone(self.staff_member.appointment_buffer_time)

        # Checking that the Config values are used
        self.assertEqual(self.staff_member.get_slot_duration(), config.slot_duration)
        self.assertEqual(self.staff_member.get_lead_time(), config.lead_time)
        self.assertEqual(self.staff_member.get_finish_time(), config.finish_time)
        self.assertEqual(self.staff_member.get_appointment_buffer_time(), config.appointment_buffer_time)

        # Setting the StaffMember values
        sm = self.get_fresh_staff_member()
        sm.slot_duration = 45
        sm.lead_time = datetime.time(9, 0)
        sm.finish_time = datetime.time(17, 0)
        sm.appointment_buffer_time = 15

        # Checking that the StaffMember values are used and not the Config values
        self.assertEqual(sm.get_slot_duration(), 45)
        self.assertEqual(sm.get_lead_time(), datetime.time(9, 0))
        self.assertEqual(sm.get_finish_time(), datetime.time(17, 0))
        self.assertEqual(sm.get_appointment_buffer_time(), 15)

    def test_get_slot_duration_and_appt_buffer_time_text(self):
        """Test get_slot_duration_text & get_appointment_buffer_time_text function."""
        sm = self.get_fresh_staff_member()
        sm.slot_duration = 33
        sm.appointment_buffer_time = 24
        self.assertEqual(sm.get_slot_duration_text(), "33 minutes")
        self.assertEqual(sm.get_appointment_buffer_time_text(), "24 minutes")



from appointment.models import PaymentInfo
from appointment.tests.base.base_test import BaseTest


class PaymentInfoBasicTestCase(BaseTest):
    @classmethod
    def setUpTestData(cls):
        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls):
        return super().tearDownClass()

    def setUp(self):
        self.ar = self.create_appt_request_for_sm1()
        self.appointment = self.create_appt_for_sm1(appointment_request=self.ar)
        self.payment_info = PaymentInfo.objects.create(appointment=self.appointment)
        return super().setUp()

    def tearDown(self):
        self.ar.delete()
        self.appointment.delete()
        self.payment_info.delete()
        return super().tearDown()

    def test_str_representation(self):
        """Test if payment info's string representation is correct."""
        self.assertEqual(str(self.payment_info), f"{self.service1.name} - {self.service1.price}")

    def test_default_attributes_on_creation(self):
        """Test if payment info can be created."""
        payment_info = PaymentInfo.objects.get(appointment=self.appointment)
        self.assertIsNotNone(payment_info)
        self.assertEqual(payment_info.appointment, self.appointment)
        self.assertEqual(self.payment_info.get_id_request(), self.appointment.get_appointment_id_request())
        self.assertEqual(self.payment_info.get_amount_to_pay(), self.appointment.get_appointment_amount_to_pay())
        self.assertEqual(self.payment_info.get_currency(), self.appointment.get_appointment_currency())
        self.assertEqual(self.payment_info.get_name(), self.appointment.get_service_name())
        self.assertIsNotNone(self.payment_info.created_at)
        self.assertIsNotNone(self.payment_info.updated_at)

    def test_get_user_info(self):
        """Test if payment info's username is correct."""
        self.assertEqual(self.payment_info.get_user_name(), self.users['client1'].first_name)
        self.assertEqual(self.payment_info.get_user_email(), self.users['client1'].email)


class PaymentInfoStatusTestCase(BaseTest):
    @classmethod
    def setUpTestData(cls):
        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls):
        return super().tearDownClass()

    def setUp(self):
        self.ar = self.create_appt_request_for_sm1()
        self.appointment = self.create_appt_for_sm1(appointment_request=self.ar)
        self.payment_info = PaymentInfo.objects.create(appointment=self.appointment)
        return super().setUp()

    def tearDown(self):
        self.ar.delete()
        self.appointment.delete()
        self.payment_info.delete()
        return super().tearDown()

    def test_set_paid_status(self):
        """Test if payment info's paid status can be set correctly."""
        self.payment_info.set_paid_status(True)
        self.assertTrue(self.appointment.is_paid())
        self.payment_info.set_paid_status(False)
        self.assertFalse(self.appointment.is_paid())



# logger_config.py
# Path: appointment/logger_config.py

"""
Author: Adams Pierre David
Since: 1.1.0
"""

import logging
import sys
from datetime import datetime

import colorama

# Initialize colorama for cross-platform color support
colorama.init()


class ColoredFormatter(logging.Formatter):
    COLORS = {
        'DEBUG': colorama.Fore.BLUE,
        'INFO': colorama.Fore.GREEN,
        'WARNING': colorama.Fore.YELLOW,
        'ERROR': colorama.Fore.RED,
        'CRITICAL': colorama.Fore.RED + colorama.Style.BRIGHT,
    }

    def format(self, record):
        log_color = self.COLORS.get(record.levelname, colorama.Fore.WHITE)
        log_time = datetime.fromtimestamp(record.created).strftime('%d/%b/%Y %H:%M:%S')

        log_msg = (
            f"{log_color}[{log_time}] {record.levelname:<4}{colorama.Style.RESET_ALL} "
            f"{colorama.Fore.LIGHTBLUE_EX}{record.name}:{record.funcName}:{record.lineno}{colorama.Style.RESET_ALL} "
            f"- {record.getMessage()}"
        )

        if record.exc_info:
            log_msg += '\n' + self.formatException(record.exc_info)
        return log_msg


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # Create a colored formatter
    formatter = ColoredFormatter()

    # Create a stream handler
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(stream_handler)

    return logger



# session.py
# Path: appointment/utils/session.py

"""
Author: Adams Pierre David
Since: 1.1.0
"""

from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import gettext as _
from phonenumber_field.phonenumber import PhoneNumber

from appointment.logger_config import get_logger
from appointment.utils.db_helpers import get_user_by_email
from appointment.utils.email_ops import send_verification_email

logger = get_logger(__name__)


def handle_existing_email(request, client_data, appointment_data, appointment_request_id, id_request):
    """
    Handle the case where the email already exists in the database.

    Sends a verification email to the existing user and redirects the client to enter the verification code.

    If the email is already in the session variables, clean the session variables for email, phone, want_reminder,
    address, and additional_info. Then, store the current email, phone, want_reminder, address, and additional_info
    in the session.

    :param request: The Django HttpRequest object.
    :param client_data: The data of the client for the appointment.
    :param appointment_data: The data of the appointment.
    :param appointment_request_id: The ID of the appointment request.
    :param id_request: The unique ID for the appointment request.
    :return: The redirect response to enter the verification code.
    """
    logger.info("Email already in database, saving info in session and redirecting to enter verification code")
    user = get_user_by_email(client_data['email'])
    send_verification_email(user=user, email=client_data['email'])

    # clean the session variables
    session_keys = ['email', 'phone', 'want_reminder', 'address', 'additional_info']
    phone = appointment_data['phone']

    for key in session_keys:
        if key in request.session:
            del request.session[key]

    request.session['email'] = client_data['email']
    request.session['phone'] = str(phone)
    request.session['want_reminder'] = appointment_data['want_reminder']
    request.session['address'] = appointment_data['address']
    request.session['additional_info'] = appointment_data['additional_info']

    # request.session['BASE_TEMPLATE'] = get_generic_context(request, admin=False)['BASE_TEMPLATE']
    request.session.modified = True
    message = _("Email '{email}' already exists. Login to your account.").format(email=client_data['email'])
    messages.error(request, message)
    return redirect('appointment:enter_verification_code', appointment_request_id=appointment_request_id,
                    id_request=id_request)


def handle_email_change(request, user, email):
    send_verification_email(user=user, email=email)
    # clean the session variables
    session_keys = ['email', 'old_email']
    for key in session_keys:
        if key in request.session:
            del request.session[key]
    request.session['email'] = email
    request.session['old_email'] = user.email
    request.session.modified = True
    return redirect('appointment:email_change_verification_code')


def get_appointment_data_from_session(request):
    """
    Get the appointment data from the session variables.
    Retrieves the phone, want_reminder, address, and additional_info stored in the session.

    :param request: The Django HttpRequest object.
    :return: The appointment data retrieved from the session.
    """
    phone = request.session.get('phone')
    phone_obj = PhoneNumber.from_string(phone)
    want_reminder = request.session.get('want_reminder') == 'on'
    address = request.session.get('address')
    additional_info = request.session.get('additional_info')

    return {
        'phone': phone_obj,
        'want_reminder': want_reminder,
        'address': address,
        'additional_info': additional_info,
    }



from copy import deepcopy
from datetime import date, datetime, time, timedelta

from django.core.exceptions import ValidationError
from django.utils import timezone

from appointment.tests.base.base_test import BaseTest


class AppointmentRequestCreationAndBasicAttributesTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls):
        return super().tearDownClass()

    def setUp(self) -> None:
        self.ar = self.create_appt_request_for_sm1()
        return super().setUp()

    def tearDown(self):
        self.ar.delete()
        super().tearDown()

    def test_default_attributes_on_creation(self):
        self.assertIsNotNone(self.ar)
        self.assertEqual(self.ar.service, self.service1)
        self.assertEqual(self.ar.staff_member, self.staff_member1)
        self.assertEqual(self.ar.start_time, time(9, 0))
        self.assertEqual(self.ar.end_time, time(10, 0))
        self.assertIsNotNone(self.ar.get_id_request())
        self.assertEqual(self.ar.date, timezone.now().date())
        self.assertTrue(isinstance(self.ar.get_id_request(), str))
        self.assertIsNotNone(self.ar.created_at)
        self.assertIsNotNone(self.ar.updated_at)

    def test_appointment_request_initial_state(self):
        """Check the initial state of "reschedule attempts" and string representation."""
        self.assertEqual(self.ar.reschedule_attempts, 0)
        expected_representation = f"{self.ar.date} - {self.ar.start_time} to {self.ar.end_time} - {self.ar.service.name}"
        self.assertEqual(str(self.ar), expected_representation)


class AppointmentRequestServiceAttributesTests(BaseTest):
    @classmethod
    def setUpTestData(cls):
        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls):
        return super().tearDownClass()

    def setUp(self) -> None:
        self.ar = self.create_appt_request_for_sm1()
        return super().setUp()

    def tearDown(self):
        self.ar.delete()
        super().tearDown()

    def test_service_related_attributes_are_correct(self):
        """Validate attributes related to the service within an appointment request."""
        self.assertEqual(self.ar.get_service_name(), self.service1.name)
        self.assertEqual(self.ar.get_service_price(), self.service1.get_price())
        self.assertEqual(self.ar.get_service_down_payment(), self.service1.get_down_payment())
        self.assertEqual(self.ar.get_service_image(), self.service1.image)
        self.assertEqual(self.ar.get_service_image_url(), self.service1.get_image_url())
        self.assertEqual(self.ar.get_service_description(), self.service1.description)
        self.assertTrue(self.ar.is_a_paid_service())
        self.assertEqual(self.ar.payment_type, 'full')
        self.assertFalse(self.ar.accepts_down_payment())


class AppointmentRequestAttributeValidation(BaseTest):
    @classmethod
    def setUpTestData(cls):
        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls):
        return super().tearDownClass()

    def setUp(self) -> None:
        self.ar = self.create_appt_request_for_sm1()
        return super().setUp()

    def tearDown(self):
        self.ar.delete()
        super().tearDown()

    def test_appointment_request_time_validations(self):
        """Ensure start and end times are validated correctly."""
        ar = deepcopy(self.ar)

        # End time before start time
        ar.start_time = time(11, 0)
        ar.end_time = time(9, 0)
        with self.assertRaises(ValidationError):
            ar.full_clean()

        # End time equal to start time
        ar.end_time = time(11, 0)
        with self.assertRaises(ValidationError):
            ar.full_clean()

        with self.assertRaises(ValidationError, msg="Start time and end time cannot be the same"):
            self.create_appointment_request_(
                self.service1, self.staff_member1, date_=date.today(), start_time=time(10, 0), end_time=time(10, 0)
            )

    def test_appointment_request_date_validations(self):
        """Validate that appointment requests cannot be in the past or have invalid durations."""
        ar = deepcopy(self.ar)

        past_date = date.today() - timedelta(days=30)
        ar.date = past_date
        with self.assertRaises(ValidationError):
            ar.full_clean()

        with self.assertRaises(ValidationError, msg="Date cannot be in the past"):
            self.create_appointment_request_(self.service1, self.staff_member1, date_=past_date)

        with self.assertRaises(ValidationError, msg="The date is not valid"):
            date_ = datetime.strptime("31-03-2021", "%d-%m-%Y").date()
            self.create_appointment_request_(
                self.service1, self.staff_member1, date_=date_)

    def test_appointment_duration_exceeds_service_time(self):
        """Test that an appointment cannot be created with a duration greater than the service duration."""
        long_duration = timedelta(hours=3)
        service = self.create_service_(name="Asgard Technology Retrofit", duration=long_duration)
        service.duration = long_duration
        service.save()

        # Create an appointment request with a 4-hour duration and the 3-hour service (should not work)
        with self.assertRaises(ValidationError):
            self.create_appointment_request_(service, self.staff_member1, start_time=time(9, 0),
                                             end_time=time(13, 0))

    def test_invalid_payment_type_raises_error(self):
        """Payment type must be either 'full' or 'down'"""
        ar = deepcopy(self.ar)
        ar.payment_type = "Naquadah Instead of Credits"
        with self.assertRaises(ValidationError):
            ar.full_clean()


class AppointmentRequestRescheduleHistory(BaseTest):
    @classmethod
    def setUpTestData(cls):
        return super().setUpTestData()

    @classmethod
    def tearDownClass(cls):
        return super().tearDownClass()

    def setUp(self) -> None:
        service = deepcopy(self.service1)
        service.reschedule_limit = 2
        service.allow_rescheduling = True
        service.save()
        self.ar_ = self.create_appt_request_for_sm1(service=service)
        return super().setUp()

    def test_ar_can_be_reschedule(self):
        self.assertTrue(self.ar_.can_be_rescheduled())

    def test_reschedule_attempts_increment(self):
        self.assertTrue(self.ar_.can_be_rescheduled())
        self.ar_.increment_reschedule_attempts()
        self.assertEqual(self.ar_.reschedule_attempts, 1)
        self.assertTrue(self.ar_.can_be_rescheduled())
        self.ar_.increment_reschedule_attempts()
        self.assertEqual(self.ar_.reschedule_attempts, 2)
        self.assertFalse(self.ar_.can_be_rescheduled())

    def test_no_reschedule_history(self):
        service = deepcopy(self.service1)
        ar = self.create_appointment_request_(service, self.staff_member1)
        self.assertFalse(ar.get_reschedule_history().exists())



from datetime import time

from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.test import TestCase, override_settings

from appointment.models import Config


class ConfigCreationTestCase(TestCase):
    def setUp(self):
        self.config = Config.objects.create(slot_duration=30, lead_time=time(9, 0),
                                            finish_time=time(17, 0), appointment_buffer_time=2.0,
                                            website_name="Stargate Command")

    @override_settings(DEBUG=True)
    def tearDown(self):
        Config.objects.all().delete()
        cache.clear()
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_default_attributes_on_creation(self):
        self.assertIsNotNone(self.config)
        self.assertEqual(self.config.slot_duration, 30)
        self.assertEqual(self.config.lead_time, time(9, 0))
        self.assertEqual(self.config.finish_time, time(17, 0))
        self.assertEqual(self.config.appointment_buffer_time, 2.0)
        self.assertEqual(self.config.website_name, "Stargate Command")
        self.assertIsNotNone(Config.get_instance())

    def test_multiple_config_creation(self):
        """Test that only one configuration can be created."""
        with self.assertRaises(ValidationError):
            Config.objects.create(slot_duration=20, lead_time=time(8, 0), finish_time=time(18, 0))

    def test_config_str_method(self):
        """Test that the string representation of a configuration is correct."""
        expected_str = f"Config {self.config.pk}: slot_duration=30, lead_time=09:00:00, finish_time=17:00:00"
        self.assertEqual(str(self.config), expected_str)


class ConfigUpdateTestCase(TestCase):
    def setUp(self):
        self.config = Config.objects.create(slot_duration=30, lead_time=time(9, 0),
                                            finish_time=time(17, 0), appointment_buffer_time=2.0,
                                            website_name="Stargate Command")

    @override_settings(DEBUG=True)
    def tearDown(self):
        Config.objects.all().delete()
        cache.clear()
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_editing_existing_config(self):
        """Test that an existing configuration can be edited."""
        self.config.slot_duration = 41
        self.config.website_name = "Cheyeene Mountain Complex"
        self.config.save()

        updated_config = Config.objects.get(pk=self.config.pk)
        self.assertEqual(updated_config.website_name, "Cheyeene Mountain Complex")
        self.assertEqual(updated_config.slot_duration, 41)


class ConfigDeletionTestCase(TestCase):
    def setUp(self):
        self.config = Config.objects.create(slot_duration=30, lead_time=time(9, 0),
                                            finish_time=time(17, 0), appointment_buffer_time=2.0,
                                            website_name="Stargate Command")

    @override_settings(DEBUG=True)
    def tearDown(self):
        Config.objects.all().delete()
        cache.clear()
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_cant_delete_config(self):
        """Test that a configuration cannot be deleted."""
        self.config.delete()
        self.assertIsNotNone(Config.objects.first())


class ConfigDurationValidationTestCase(TestCase):
    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_invalid_slot_duration(self):
        """Test that a configuration cannot be created with a negative slot duration."""
        with self.assertRaises(ValidationError):
            Config.objects.create(slot_duration=-10, lead_time=time(9, 0), finish_time=time(17, 0))
        with self.assertRaises(ValidationError):
            Config.objects.create(slot_duration=0, lead_time=time(9, 0), finish_time=time(17, 0))


class ConfigTimeValidationTestCase(TestCase):
    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def test_lead_time_greater_than_finish_time(self):
        # TODO: Think about business with night shifts, start time will be greater than finish time,
        #  but again, not sure client will use this app for night shifts
        """Test that lead time cannot be greater than finish time."""
        with self.assertRaises(ValidationError):
            Config.objects.create(slot_duration=30, lead_time=time(18, 0), finish_time=time(9, 0))

    def test_same_lead_and_finish_time(self):
        """Test that a configuration cannot be created with the same lead time and finish time."""
        with self.assertRaises(ValidationError):
            Config.objects.create(slot_duration=30, lead_time=time(9, 0), finish_time=time(9, 0))

    def test_negative_appointment_buffer_time(self):
        """Test that a configuration cannot be created with a negative appointment buffer time."""
        with self.assertRaises(ValidationError):
            Config.objects.create(slot_duration=30, lead_time=time(9, 0), finish_time=time(17, 0),
                                  appointment_buffer_time=-2.0)



from copy import deepcopy
from datetime import datetime, time, timedelta

from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.translation import gettext as _

from appointment.models import Appointment, DayOff, WorkingHours
from appointment.tests.base.base_test import BaseTest
from appointment.utils.date_time import get_weekday_num


class AppointmentCreationTestCase(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        self.address = 'Stargate Command, Cheyenne Mountain Complex, Colorado Springs, CO'
        self.ar = self.create_appt_request_for_sm1()
        self.appointment = self.create_appt_for_sm1(appointment_request=self.ar)
        self.client_ = self.users['client1']
        self.expected_end_time = datetime.combine(self.ar.date, self.ar.end_time)
        self.expected_service_name = 'Stargate Activation'
        self.expected_service_price = 100000
        self.expected_start_time = datetime.combine(self.ar.date, self.ar.start_time)
        self.phone = '+12392340543'
        return super().setUp()

    def tearDown(self):
        self.appointment.delete()
        return super().tearDown()

    def test_default_attributes_on_creation(self):
        """Test default attributes when an appointment is created."""
        self.assertIsNotNone(self.appointment)
        self.assertIsNotNone(self.appointment.created_at)
        self.assertIsNotNone(self.appointment.updated_at)
        self.assertIsNotNone(self.appointment.get_appointment_id_request())
        self.assertIsNone(self.appointment.additional_info)
        self.assertEqual(self.appointment.client, self.users['client1'])
        self.assertEqual(self.appointment.phone, self.phone)
        self.assertEqual(self.appointment.address, self.address)
        self.assertFalse(self.appointment.want_reminder)

    def test_str_representation(self):
        """Test if an appointment's string representation is correct."""
        expected_str = f"{self.client_} - {self.ar.start_time.strftime('%Y-%m-%d %H:%M')} to " \
                       f"{self.ar.end_time.strftime('%Y-%m-%d %H:%M')}"
        self.assertEqual(str(self.appointment), expected_str)

    def test_appointment_getters(self):
        """Test getter methods for appointment details."""
        self.assertEqual(self.appointment.get_start_time(), self.expected_start_time)
        self.assertEqual(self.appointment.get_end_time(), self.expected_end_time)
        self.assertEqual(self.appointment.get_service_name(), self.expected_service_name)
        self.assertEqual(self.appointment.get_service_price(), self.expected_service_price)
        self.assertEqual(self.appointment.is_paid_text(), _('No'))
        self.assertEqual(self.appointment.get_appointment_amount_to_pay(), self.expected_service_price)
        self.assertEqual(self.appointment.get_service_down_payment(), self.service1.get_down_payment())
        self.assertEqual(self.appointment.get_service_description(), self.service1.description)
        self.assertEqual(self.appointment.get_appointment_date(), self.ar.date)
        self.assertEqual(self.appointment.get_service_duration(), "1 hour")
        self.assertEqual(self.appointment.get_appointment_currency(), "USD")
        self.assertEqual(self.appointment.get_appointment_amount_to_pay(), self.ar.get_service_price())
        self.assertEqual(self.appointment.get_service_img_url(), "")
        self.assertEqual(self.appointment.get_staff_member_name(), self.staff_member1.get_staff_member_name())
        self.assertTrue(self.appointment.service_is_paid())
        self.assertFalse(self.appointment.is_paid())

    def test_conversion_to_dict(self):
        response = {
            'id': 1,
            'client_name': self.client_.first_name + ' ' + self.client_.last_name,
            'client_email': self.client_.email,
            'start_time': '1900-01-01 09:00',
            'end_time': '1900-01-01 10:00',
            'service_name': self.expected_service_name,
            'address': self.address,
            'want_reminder': False,
            'additional_info': None,
            'paid': False,
            'amount_to_pay': self.expected_service_price,
        }
        actual_response = self.appointment.to_dict()
        actual_response.pop('id_request', None)
        self.assertEqual(actual_response, response)


class AppointmentValidDateTestCase(BaseTest):
    def setUp(self):
        super().setUp()
        self.weekday = "Monday"  # Example weekday
        self.weekday_num = get_weekday_num(self.weekday)
        self.wh = WorkingHours.objects.create(staff_member=self.staff_member1, day_of_week=self.weekday_num,
                                              start_time=time(9, 0), end_time=time(17, 0))
        self.appt_date = timezone.now().date() + timedelta(days=(self.weekday_num - timezone.now().weekday()) % 7)
        self.start_time = timezone.now().replace(hour=10, minute=0, second=0, microsecond=0)
        self.current_appointment_id = None

    def tearDown(self):
        self.wh.delete()
        return super().tearDown()

    def test_staff_member_works_on_given_day(self):
        is_valid, message = Appointment.is_valid_date(self.appt_date, self.start_time, self.staff_member1,
                                                      self.current_appointment_id, self.weekday)
        self.assertTrue(is_valid)

    def test_staff_member_does_not_work_on_given_day(self):
        non_working_day = "Sunday"
        non_working_day_num = get_weekday_num(non_working_day)
        appt_date = self.appt_date + timedelta(days=(non_working_day_num - self.weekday_num) % 7)
        is_valid, message = Appointment.is_valid_date(appt_date, self.start_time, self.staff_member1,
                                                      self.current_appointment_id, non_working_day)
        self.assertFalse(is_valid)
        self.assertIn("does not work on this day", message)

    def test_start_time_outside_working_hours(self):
        early_start_time = timezone.now().replace(hour=8, minute=0)  # Before working hours
        is_valid, message = Appointment.is_valid_date(self.appt_date, early_start_time, self.staff_member1,
                                                      self.current_appointment_id, self.weekday)
        self.assertFalse(is_valid)
        self.assertIn("outside of", message)

    def test_staff_member_has_day_off(self):
        DayOff.objects.create(staff_member=self.staff_member1, start_date=self.appt_date, end_date=self.appt_date)
        is_valid, message = Appointment.is_valid_date(self.appt_date, self.start_time, self.staff_member1,
                                                      self.current_appointment_id, self.weekday)
        self.assertFalse(is_valid)
        self.assertIn("has a day off on this date", message)


class AppointmentValidationTestCase(BaseTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        self.tomorrow = timezone.now().date() + timedelta(days=1)
        self.ar = self.create_appointment_request_(service=self.service2, staff_member=self.staff_member2,
                                                   date_=self.tomorrow)
        self.appointment = self.create_appt_for_sm2(appointment_request=self.ar)
        self.client_ = self.users['client1']
        return super().setUp()

    def tearDown(self):
        self.appointment.delete()
        return super().tearDown()

    def test_invalid_phone_number(self):
        """Test that an appointment cannot be created with an invalid phone number."""
        self.appointment.phone = "1234"  # Invalid phone number
        with self.assertRaises(ValidationError):
            self.appointment.full_clean()

    def test_set_paid_status(self):
        """Test if an appointment's paid status can be set."""
        appointment = deepcopy(self.appointment)
        appointment.set_appointment_paid_status(True)
        self.assertTrue(appointment.is_paid())
        appointment.set_appointment_paid_status(False)
        self.assertFalse(appointment.is_paid())

    def test_save_with_down_payment(self):
        """Test if an appointment can be saved with a down payment."""
        ar = self.create_appointment_request_(service=self.service2, staff_member=self.staff_member2,
                                              date_=self.tomorrow)
        appointment = self.create_appt_for_sm2(appointment_request=ar)
        ar.payment_type = 'down'
        ar.save()
        appointment.save()
        self.assertEqual(appointment.get_service_down_payment(), self.service2.get_down_payment())

    def test_creation_without_appointment_request(self):
        """Test that an appointment cannot be created without an appointment request."""
        with self.assertRaises(ValidationError):  # Assuming model validation prevents this
            Appointment.objects.create(client=self.client_)

    def test_creation_without_client(self):
        """Test that an appointment can be created without a client."""
        ar = self.create_appointment_request_(service=self.service2, staff_member=self.staff_member2,
                                              date_=self.tomorrow)
        appt = Appointment.objects.create(appointment_request=ar)
        self.assertIsNone(appt.client)

    def test_creation_without_required_fields(self):
        """Test that an appointment cannot be created without the required fields."""
        with self.assertRaises(ValidationError):
            Appointment.objects.create()

    def test_get_staff_member_name_without_staff_member(self):
        """Test if you get_staff_member_name method returns an empty string when no staff member is associated."""
        ar = self.create_appointment_request_(service=self.service2, staff_member=self.staff_member2,
                                              date_=self.tomorrow)
        appointment = self.create_appt_for_sm2(appointment_request=ar)
        appointment.appointment_request.staff_member = None
        appointment.appointment_request.save()
        self.assertEqual(appointment.get_staff_member_name(), "")

    def test_rescheduling(self):
        """Simulate appointment rescheduling by changing the appointment date and times."""
        ar = self.create_appointment_request_(service=self.service2, staff_member=self.staff_member2,
                                              date_=self.tomorrow)
        appointment = self.create_appt_for_sm2(appointment_request=ar)
        new_date = ar.date + timedelta(days=1)
        new_start_time = time(10, 0)
        new_end_time = time(11, 0)
        ar.date = new_date
        ar.start_time = new_start_time
        ar.end_time = new_end_time
        ar.save()

        self.assertEqual(appointment.get_date(), new_date)
        self.assertEqual(appointment.get_start_time().time(), new_start_time)
        self.assertEqual(appointment.get_end_time().time(), new_end_time)



from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def not_in_the_past(date):
    if date < date.today():
        raise ValidationError(_('Date is in the past'))
